# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________


import sys
import logging
import itertools

from pyomo.common.numeric_types import native_types, native_numeric_types
from pyomo.core.base import Constraint, Objective, ComponentMap

import pyomo.core.expr as EXPR
from pyomo.core.expr.numvalue import NumericConstant
from pyomo.core.base.objective import ObjectiveData, ScalarObjective
from pyomo.core.base import Expression
from pyomo.core.base.expression import (
    ScalarExpression,
    NamedExpressionData,
    ExpressionData,
)
from pyomo.core.base.var import ScalarVar, Var, VarData, value
from pyomo.core.base.param import ScalarParam, ParamData
from pyomo.core.kernel.expression import expression, noclone
from pyomo.core.kernel.variable import IVariable, variable
from pyomo.core.kernel.objective import objective

from io import StringIO

logger = logging.getLogger('pyomo.core')


#
# This checks if the first argument is a numeric value.  If not
# then this is a Pyomo constant expression, and we can only check if its
# close to 'b' when it is constant.
#
def isclose_const(a, b, rel_tol=1e-9, abs_tol=0.0):
    if not a.__class__ in native_numeric_types:
        if a.is_constant():
            a = value(a)
        else:
            return False
    return abs(a - b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)


class StandardRepn:
    """
    This class defines a standard/common representation for Pyomo expressions
    that provides an efficient interface for writing all models.

    TODO: define what "efficient" means to us.
    """

    __slots__ = (
        'constant',  # The constant term
        'linear_coefs',  # Linear coefficients
        'linear_vars',  # Linear variables
        'quadratic_coefs',  # Quadratic coefficients
        'quadratic_vars',  # Quadratic variables
        'nonlinear_expr',  # Nonlinear expression
        'nonlinear_vars',
    )  # Variables that appear in the nonlinear expression

    def __init__(self):
        self.constant = 0
        self.linear_vars = tuple()
        self.linear_coefs = tuple()
        self.quadratic_vars = tuple()
        self.quadratic_coefs = tuple()
        self.nonlinear_expr = None
        self.nonlinear_vars = tuple()

    def __getstate__(self):
        """
        This method is required because this class uses slots.
        """
        return (
            self.constant,
            self.linear_coefs,
            self.linear_vars,
            self.quadratic_coefs,
            self.quadratic_vars,
            self.nonlinear_expr,
            self.nonlinear_vars,
        )

    def __setstate__(self, state):
        """
        This method is required because this class uses slots.
        """
        (
            self.constant,
            self.linear_coefs,
            self.linear_vars,
            self.quadratic_coefs,
            self.quadratic_vars,
            self.nonlinear_expr,
            self.nonlinear_vars,
        ) = state

    #
    # Generate a string representation of the expression
    #
    def __str__(self):  # pragma: nocover
        output = StringIO()
        output.write("\n")
        output.write("constant:       " + str(self.constant) + "\n")
        output.write(
            "linear vars:    " + str([v_.name for v_ in self.linear_vars]) + "\n"
        )
        output.write(
            "linear var ids: " + str([id(v_) for v_ in self.linear_vars]) + "\n"
        )
        output.write("linear coef:    " + str(list(self.linear_coefs)) + "\n")
        output.write(
            "quadratic vars:    "
            + str([(v_[0].name, v_[1].name) for v_ in self.quadratic_vars])
            + "\n"
        )
        output.write(
            "quadratic var ids: "
            + str([(id(v_[0]), id(v_[1])) for v_ in self.quadratic_vars])
            + "\n"
        )
        output.write("quadratic coef:    " + str(list(self.quadratic_coefs)) + "\n")
        if self.nonlinear_expr is None:
            output.write("nonlinear expr: None\n")
        else:
            output.write("nonlinear expr:\n")
            try:
                output.write(self.nonlinear_expr.to_string())
                output.write("\n")
            except AttributeError:
                output.write(str(self.nonlinear_expr))
                output.write("\n")
        output.write(
            "nonlinear vars: " + str([v_.name for v_ in self.nonlinear_vars]) + "\n"
        )
        output.write("\n")

        ret_str = output.getvalue()
        output.close()
        return ret_str

    def is_fixed(self):
        if (
            len(self.linear_vars) == 0
            and len(self.nonlinear_vars) == 0
            and len(self.quadratic_vars) == 0
        ):
            return True
        return False

    def polynomial_degree(self):
        if not self.nonlinear_expr is None:
            return None
        if len(self.quadratic_coefs) > 0:
            return 2
        if len(self.linear_coefs) > 0:
            return 1
        return 0

    def is_constant(self):
        return (
            self.nonlinear_expr is None
            and len(self.quadratic_coefs) == 0
            and len(self.linear_coefs) == 0
        )

    def is_linear(self):
        return self.nonlinear_expr is None and len(self.quadratic_coefs) == 0

    def is_quadratic(self):
        return len(self.quadratic_coefs) > 0 and self.nonlinear_expr is None

    def is_nonlinear(self):
        return not (self.nonlinear_expr is None and len(self.quadratic_coefs) == 0)

    def to_expression(self, sort=True):
        #
        # When an standard representation is created, the ordering of the
        # linear and quadratic terms may not be preserved.  Hence, the
        # sorting option ensures that an expression generated from a
        # standard representation has a consistent order.
        #
        # TODO: Should this replace non-mutable parameters with constants?
        #
        expr = self.constant

        lvars = [(i, v) for i, v in enumerate(self.linear_vars)]
        if sort:
            lvars = sorted(lvars, key=lambda x: str(x[1]))
        for i, v in lvars:
            c = self.linear_coefs[i]
            if c.__class__ in native_numeric_types:
                if not c:
                    pass
                if isclose_const(c, 1.0):
                    expr += v
                elif isclose_const(c, -1.0):
                    expr -= v
                elif c < 0.0:
                    expr -= -c * v
                else:
                    expr += c * v
            else:
                expr += c * v

        qvars = [(i, v) for i, v in enumerate(self.quadratic_vars)]
        if sort:
            qvars = sorted(qvars, key=lambda x: (str(x[1][0]), str(x[1][1])))
        for i, v in qvars:
            if id(v[0]) == id(v[1]):
                term = v[0] ** 2
            else:
                term = v[0] * v[1]
            c = self.quadratic_coefs[i]
            if c.__class__ in native_numeric_types:
                if isclose_const(c, 1.0):
                    expr += term
                elif isclose_const(c, -1.0):
                    expr -= term
                else:
                    expr += c * term
            else:
                expr += c * term

        if self.nonlinear_expr is not None:
            if expr.__class__ in native_numeric_types and expr == 0:
                # Some "NL" expressions do not support addition
                # (e.g. relational expressions)
                return self.nonlinear_expr
            expr += self.nonlinear_expr
        return expr


"""

Note:  This function separates linear terms from nonlinear terms.
Along the way, fixed variable and mutable parameter values *may* be
replaced with constants.  However, that is not guaranteed.  Thus,
the nonlinear expression may contain subexpressions whose value is
constant.  This was done to avoid additional work when a subexpression
is clearly nonlinear.  However, this requires that standard
representations be temporary.  They should be used to interface
to a solver and then be deleted.

"""


# @profile
def generate_standard_repn(
    expr, idMap=None, compute_values=True, verbose=False, quadratic=True, repn=None
):
    #
    # Use a custom Results object
    #
    global Results
    if quadratic:
        Results = ResultsWithQuadratics
    else:
        Results = ResultsWithoutQuadratics

    if True:
        #
        # Setup
        #
        if idMap is None:
            idMap = {}
        idMap.setdefault(None, {})
        if repn is None:
            repn = StandardRepn()
        #
        # The expression is a number or a non-variable constant
        # expression.
        #
        if expr.__class__ in native_numeric_types or not expr.is_potentially_variable():
            if compute_values:
                repn.constant = EXPR.evaluate_expression(expr)
            else:
                repn.constant = expr
            return repn
        #
        # The expression is a variable
        #
        elif expr.is_variable_type():
            if expr.fixed:
                if compute_values:
                    repn.constant = value(expr)
                else:
                    repn.constant = expr
                return repn
            repn.linear_coefs = (1,)
            repn.linear_vars = (expr,)
            return repn
        #
        # The expression is linear
        #
        elif expr.__class__ is EXPR.LinearExpression:
            linear_coefs = {}
            linear_vars = {}
            C_ = 0
            if compute_values:
                for arg in expr.args:
                    if arg.__class__ is EXPR.MonomialTermExpression:
                        c, v = arg.args
                        if c.__class__ not in native_numeric_types:
                            c = EXPR.evaluate_expression(c)
                        if v.fixed:
                            C_ += c * v.value
                            continue
                        id_ = id(v)
                        if id_ in linear_coefs:
                            linear_coefs[id_] += c
                        else:
                            linear_coefs[id_] = c
                            linear_vars[id_] = v
                    elif arg.__class__ in native_numeric_types:
                        C_ += arg
                    elif arg.is_variable_type():
                        if arg.fixed:
                            C_ += arg.value
                            continue
                        id_ = id(arg)
                        if id_ in linear_coefs:
                            linear_coefs[id_] += 1
                        else:
                            linear_coefs[id_] = 1
                            linear_vars[id_] = arg
                    else:
                        C_ += EXPR.evaluate_expression(arg)
            else:  # compute_values == False
                for arg in expr.args:
                    if arg.__class__ is EXPR.MonomialTermExpression:
                        c, v = arg.args
                        if v.fixed:
                            C_ += c * v
                            continue
                        id_ = id(v)
                        if id_ in linear_coefs:
                            linear_coefs[id_] += c
                        else:
                            linear_coefs[id_] = c
                            linear_vars[id_] = v
                    elif arg.__class__ in native_numeric_types:
                        C_ += arg
                    elif arg.is_variable_type():
                        if arg.fixed:
                            C_ += arg
                            continue
                        id_ = id(arg)
                        if id_ in linear_coefs:
                            linear_coefs[id_] += 1
                        else:
                            linear_coefs[id_] = 1
                            linear_vars[id_] = arg
                    else:
                        C_ += arg

            vars_ = []
            coef_ = []
            for id_, coef in linear_coefs.items():
                if coef.__class__ in native_numeric_types and not coef:
                    continue
                if id_ not in idMap[None]:
                    key = len(idMap) - 1
                    idMap[None][id_] = key
                    idMap[key] = linear_vars[id_]
                else:
                    key = idMap[None][id_]
                vars_.append(idMap[key])
                coef_.append(coef)

            repn.linear_vars = tuple(vars_)
            repn.linear_coefs = tuple(coef_)
            repn.constant = C_
            return repn

        #
        # Unknown expression object
        #
        elif not expr.is_expression_type():  # pragma: nocover
            raise ValueError("Unexpected expression type: " + str(expr))

        #
        # WEH - Checking the polynomial degree didn't
        #       turn out to be a win.  But I'm leaving this
        #       in as a comment for now, since we're not
        #       done tuning this code.
        #
        # degree = expr.polynomial_degree()
        # if degree == 1:
        #    return _generate_linear_standard_repn(expr,
        #                        idMap=idMap,
        #                        compute_values=compute_values,
        #                        verbose=verbose,
        #                        repn=repn)
        # else:
        return _generate_standard_repn(
            expr,
            idMap=idMap,
            compute_values=compute_values,
            verbose=verbose,
            quadratic=quadratic,
            repn=repn,
        )


##-----------------------------------------------------------------------
##
## Logic for _generate_standard_repn
##
##-----------------------------------------------------------------------


class ResultsWithQuadratics:
    __slot__ = ('const', 'nonl', 'linear', 'quadratic')

    def __init__(self, constant=0, nonl=0, linear=None, quadratic=None):
        self.constant = constant
        self.nonl = nonl
        self.linear = {}
        # if linear is None:
        #    self.linear = {}
        # else:
        #    self.linear = linear
        self.quadratic = {}
        # if quadratic is None:
        #    self.quadratic = {}
        # else:
        #    self.quadratic = quadratic

    def __str__(self):  # pragma: nocover
        return "Const:\t%s\nLinear:\t%s\nQuadratic:\t%s\nNonlinear:\t%s" % (
            str(self.constant),
            str(self.linear),
            str(self.quadratic),
            str(self.nonl),
        )


class ResultsWithoutQuadratics:
    __slot__ = ('const', 'nonl', 'linear')

    def __init__(self, constant=0, nonl=0, linear=None):
        self.constant = constant
        self.nonl = nonl
        self.linear = {}
        # if linear is None:
        #    self.linear = {}
        # else:
        #    self.linear = linear

    def __str__(self):  # pragma: nocover
        return "Const:\t%s\nLinear:\t%s\nNonlinear:\t%s" % (
            str(self.constant),
            str(self.linear),
            str(self.nonl),
        )


Results = ResultsWithQuadratics


# @profile
def _collect_sum(exp, multiplier, idMap, compute_values, verbose, quadratic):
    ans = Results()
    nonl = []
    varkeys = idMap[None]

    for e_ in itertools.islice(exp._args_, exp.nargs()):
        if e_.__class__ is EXPR.MonomialTermExpression:
            lhs, v = e_.args
            if lhs.__class__ in native_numeric_types:
                if not lhs:
                    continue
            elif compute_values:
                lhs = value(lhs)
                if not lhs:
                    continue
            if v.fixed:
                if compute_values:
                    ans.constant += multiplier * lhs * value(v)
                else:
                    ans.constant += multiplier * lhs * v
            else:
                id_ = id(v)
                if id_ in varkeys:
                    key = varkeys[id_]
                else:
                    key = len(idMap) - 1
                    varkeys[id_] = key
                    idMap[key] = v
                if key in ans.linear:
                    ans.linear[key] += multiplier * lhs
                else:
                    ans.linear[key] = multiplier * lhs
        elif e_.__class__ in native_numeric_types:
            ans.constant += multiplier * e_
        elif e_.is_variable_type():
            if e_.fixed:
                if compute_values:
                    ans.constant += multiplier * e_.value
                else:
                    ans.constant += multiplier * e_
            else:
                id_ = id(e_)
                if id_ in varkeys:
                    key = varkeys[id_]
                else:
                    key = len(idMap) - 1
                    varkeys[id_] = key
                    idMap[key] = e_
                if key in ans.linear:
                    ans.linear[key] += multiplier
                else:
                    ans.linear[key] = multiplier
        elif not e_.is_potentially_variable():
            if compute_values:
                ans.constant += multiplier * value(e_)
            else:
                ans.constant += multiplier * e_
        else:
            res_ = _collect_standard_repn(
                e_, multiplier, idMap, compute_values, verbose, quadratic
            )
            #
            # Add returned from recursion
            #
            ans.constant += res_.constant
            if not (res_.nonl.__class__ in native_numeric_types and res_.nonl == 0):
                nonl.append(res_.nonl)
            for i, v in res_.linear.items():
                ans.linear[i] = ans.linear.get(i, 0) + v
            if quadratic:
                for i in res_.quadratic:
                    ans.quadratic[i] = ans.quadratic.get(i, 0) + res_.quadratic[i]

    if len(nonl) > 0:
        if len(nonl) == 1:
            ans.nonl = nonl[0]
        else:
            ans.nonl = EXPR.SumExpression(nonl)
    zero_coef = [
        k
        for k, coef in ans.linear.items()
        if coef.__class__ in native_numeric_types and not coef
    ]
    for k in zero_coef:
        ans.linear.pop(k)
    return ans


# @profile
def _collect_term(exp, multiplier, idMap, compute_values, verbose, quadratic):
    #
    # LHS is a numeric value
    #
    if exp._args_[0].__class__ in native_numeric_types:
        if exp._args_[0] == 0:  # TODO: coverage?
            return Results()
        return _collect_standard_repn(
            exp._args_[1],
            multiplier * exp._args_[0],
            idMap,
            compute_values,
            verbose,
            quadratic,
        )
    #
    # LHS is a non-variable expression
    #
    else:
        if compute_values:
            val = value(exp._args_[0])
            if val == 0:
                return Results()
            return _collect_standard_repn(
                exp._args_[1],
                multiplier * val,
                idMap,
                compute_values,
                verbose,
                quadratic,
            )
        else:
            return _collect_standard_repn(
                exp._args_[1],
                multiplier * exp._args_[0],
                idMap,
                compute_values,
                verbose,
                quadratic,
            )


def _collect_prod(exp, multiplier, idMap, compute_values, verbose, quadratic):
    #
    # LHS is a numeric value
    #
    if exp._args_[0].__class__ in native_numeric_types:
        if exp._args_[0] == 0:  # TODO: coverage?
            return Results()
        return _collect_standard_repn(
            exp._args_[1],
            multiplier * exp._args_[0],
            idMap,
            compute_values,
            verbose,
            quadratic,
        )
    #
    # RHS is a numeric value
    #
    if exp._args_[1].__class__ in native_numeric_types:
        if exp._args_[1] == 0:  # TODO: coverage?
            return Results()
        return _collect_standard_repn(
            exp._args_[0],
            multiplier * exp._args_[1],
            idMap,
            compute_values,
            verbose,
            quadratic,
        )
    #
    # LHS is a non-variable expression
    #
    elif not exp._args_[0].is_potentially_variable():
        if compute_values:
            val = value(exp._args_[0])
            if val == 0:
                return Results()
            return _collect_standard_repn(
                exp._args_[1],
                multiplier * val,
                idMap,
                compute_values,
                verbose,
                quadratic,
            )
        else:
            return _collect_standard_repn(
                exp._args_[1],
                multiplier * exp._args_[0],
                idMap,
                compute_values,
                verbose,
                quadratic,
            )
    #
    # RHS is a non-variable expression
    #
    elif not exp._args_[1].is_potentially_variable():
        if compute_values:
            val = value(exp._args_[1])
            if val == 0:
                return Results()
            return _collect_standard_repn(
                exp._args_[0],
                multiplier * val,
                idMap,
                compute_values,
                verbose,
                quadratic,
            )
        else:
            return _collect_standard_repn(
                exp._args_[0],
                multiplier * exp._args_[1],
                idMap,
                compute_values,
                verbose,
                quadratic,
            )
    #
    # Both the LHS and RHS are potentially variable ...
    #
    # Collect LHS
    #
    lhs = _collect_standard_repn(
        exp._args_[0], 1, idMap, compute_values, verbose, quadratic
    )
    lhs_nonl_None = lhs.nonl.__class__ in native_numeric_types and not lhs.nonl
    #
    # LHS is potentially variable, but it turns out to be a constant
    # because the variables were fixed.
    #
    if (
        lhs_nonl_None
        and len(lhs.linear) == 0
        and (not quadratic or len(lhs.quadratic) == 0)
    ):
        if lhs.constant.__class__ in native_numeric_types and lhs.constant == 0:
            return Results()
        if compute_values:
            val = value(lhs.constant)
            if val == 0:  # TODO: coverage?
                return Results()
            return _collect_standard_repn(
                exp._args_[1],
                multiplier * val,
                idMap,
                compute_values,
                verbose,
                quadratic,
            )
        else:
            return _collect_standard_repn(
                exp._args_[1],
                multiplier * lhs.constant,
                idMap,
                compute_values,
                verbose,
                quadratic,
            )
    #
    # Collect RHS
    #
    rhs = _collect_standard_repn(
        exp._args_[1], 1, idMap, compute_values, verbose, quadratic
    )
    rhs_nonl_None = rhs.nonl.__class__ in native_numeric_types and not rhs.nonl
    #
    # If RHS is zero, then return an empty results
    #
    if (
        rhs_nonl_None
        and len(rhs.linear) == 0
        and (not quadratic or len(rhs.quadratic) == 0)
        and rhs.constant.__class__ in native_numeric_types
        and rhs.constant == 0
    ):
        return Results()
    #
    # If either the LHS or RHS are nonlinear, then simply return the nonlinear expression
    #
    if not lhs_nonl_None or not rhs_nonl_None:
        return Results(nonl=multiplier * exp)

    # If the resulting expression has a polynomial degree greater than 2
    # (1 if quadratic is False), then simply return this as a general
    # nonlinear expression
    #
    if max(1 if lhs.linear else 0, 2 if quadratic and lhs.quadratic else 0) + max(
        1 if rhs.linear else 0, 2 if quadratic and rhs.quadratic else 0
    ) > (2 if quadratic else 1):
        return Results(nonl=multiplier * exp)

    ans = Results()
    ans.constant = multiplier * lhs.constant * rhs.constant
    if not (lhs.constant.__class__ in native_numeric_types and lhs.constant == 0):
        for key, coef in rhs.linear.items():
            ans.linear[key] = multiplier * coef * lhs.constant
    if not (rhs.constant.__class__ in native_numeric_types and rhs.constant == 0):
        for key, coef in lhs.linear.items():
            if key in ans.linear:
                ans.linear[key] += multiplier * coef * rhs.constant
            else:
                ans.linear[key] = multiplier * coef * rhs.constant

    if quadratic:
        if not (lhs.constant.__class__ in native_numeric_types and lhs.constant == 0):
            for key, coef in rhs.quadratic.items():
                ans.quadratic[key] = multiplier * coef * lhs.constant
        if not (rhs.constant.__class__ in native_numeric_types and rhs.constant == 0):
            for key, coef in lhs.quadratic.items():
                if key in ans.quadratic:
                    ans.quadratic[key] += multiplier * coef * rhs.constant
                else:
                    ans.quadratic[key] = multiplier * coef * rhs.constant
        for lkey, lcoef in lhs.linear.items():
            for rkey, rcoef in rhs.linear.items():
                ndx = (lkey, rkey) if lkey <= rkey else (rkey, lkey)
                if ndx in ans.quadratic:
                    ans.quadratic[ndx] += multiplier * lcoef * rcoef
                else:
                    ans.quadratic[ndx] = multiplier * lcoef * rcoef
        # TODO - Use quicksum here?
        el_linear = multiplier * sum(
            coef * idMap[key]
            for key, coef in lhs.linear.items()
            if coef.__class__ not in native_numeric_types or coef
        )
        er_linear = multiplier * sum(
            coef * idMap[key]
            for key, coef in rhs.linear.items()
            if coef.__class__ not in native_numeric_types or coef
        )
        el_quadratic = multiplier * sum(
            coef * idMap[key[0]] * idMap[key[1]]
            for key, coef in lhs.quadratic.items()
            if coef.__class__ not in native_numeric_types or coef
        )
        er_quadratic = multiplier * sum(
            coef * idMap[key[0]] * idMap[key[1]]
            for key, coef in rhs.quadratic.items()
            if coef.__class__ not in native_numeric_types or coef
        )
        if (el_linear.__class__ not in native_numeric_types or el_linear) and (
            er_quadratic.__class__ not in native_numeric_types or er_quadratic
        ):
            ans.nonl += el_linear * er_quadratic
        if (er_linear.__class__ not in native_numeric_types or er_linear) and (
            el_quadratic.__class__ not in native_numeric_types or el_quadratic
        ):
            ans.nonl += er_linear * el_quadratic

    return ans


# @profile
def _collect_var(exp, multiplier, idMap, compute_values, verbose, quadratic):
    ans = Results()

    if exp.fixed:
        if compute_values:
            ans.constant += multiplier * value(exp)
        else:
            ans.constant += multiplier * exp
    else:
        id_ = id(exp)
        if id_ in idMap[None]:
            key = idMap[None][id_]
        else:
            key = len(idMap) - 1
            idMap[None][id_] = key
            idMap[key] = exp
        ans.linear[key] = multiplier

    return ans


def _collect_pow(exp, multiplier, idMap, compute_values, verbose, quadratic):
    #
    # Exponent is a numeric value
    #
    if exp._args_[1].__class__ in native_numeric_types:
        exponent = exp._args_[1]
    #
    # Exponent is not potentially variable
    #
    # Compute its value if compute_values==True
    #
    elif not exp._args_[1].is_potentially_variable():
        if compute_values:
            exponent = value(exp._args_[1])
        else:
            exponent = exp._args_[1]
    #
    # Otherwise collect a standard repn
    #
    else:
        res = _collect_standard_repn(
            exp._args_[1], 1, idMap, compute_values, verbose, quadratic
        )
        #
        # If the expression is variable, then return a nonlinear expression
        #
        if (
            not (res.nonl.__class__ in native_numeric_types and res.nonl == 0)
            or len(res.linear) > 0
            or (quadratic and len(res.quadratic) > 0)
        ):
            return Results(nonl=multiplier * exp)
        exponent = res.constant

    if exponent.__class__ in native_numeric_types:
        #
        # #**0 = 1
        #
        if exponent == 0:
            return Results(constant=multiplier)
        #
        # #**1 = #
        #
        # Return the standard repn for arg(0)
        #
        elif exponent == 1:
            return _collect_standard_repn(
                exp._args_[0], multiplier, idMap, compute_values, verbose, quadratic
            )
        #
        # Ignore #**2 unless quadratic==True
        #
        elif exponent == 2 and quadratic:
            res = _collect_standard_repn(
                exp._args_[0], 1, idMap, compute_values, verbose, quadratic
            )
            #
            # If arg(0) is nonlinear, then this is a nonlinear repn
            #
            if (
                not (res.nonl.__class__ in native_numeric_types and res.nonl == 0)
                or len(res.quadratic) > 0
            ):
                return Results(nonl=multiplier * exp)
            #
            # If computing values and no linear terms, then the return a constant repn
            #
            elif compute_values and len(res.linear) == 0:
                return Results(constant=multiplier * res.constant**exponent)
            #
            # If the base is linear, then we compute the quadratic expression for it.
            #
            else:
                ans = Results()
                has_constant = (
                    res.constant.__class__ not in native_numeric_types
                    or res.constant != 0
                )
                if has_constant:
                    ans.constant = multiplier * res.constant * res.constant

                # this is reversed since we want to pop off the end for efficiency
                # and the quadratic terms have a convention that the indexing tuple
                # of key1, key2 is such that key1 <= key2
                keys = sorted(res.linear.keys(), reverse=True)
                while len(keys) > 0:
                    key1 = keys.pop()
                    coef1 = res.linear[key1]
                    if has_constant:
                        ans.linear[key1] = 2 * multiplier * coef1 * res.constant
                    ans.quadratic[key1, key1] = multiplier * coef1 * coef1
                    for key2 in keys:
                        coef2 = res.linear[key2]
                        ans.quadratic[key1, key2] = 2 * multiplier * coef1 * coef2
                return ans

    #
    # If args(0) is a numeric value or it is fixed, then we have a constant value
    #
    if exp._args_[0].__class__ in native_numeric_types or exp._args_[0].is_fixed():
        if compute_values:
            return Results(constant=multiplier * value(exp._args_[0]) ** exponent)
        else:
            return Results(constant=multiplier * exp)
    #
    # Return a nonlinear expression here
    #
    return Results(nonl=multiplier * exp)


def _collect_division(exp, multiplier, idMap, compute_values, verbose, quadratic):
    if (
        exp._args_[1].__class__ in native_numeric_types
        or not exp._args_[1].is_potentially_variable()
    ):  # TODO: coverage?
        # Denominator is trivially constant
        if compute_values:
            denom = 1.0 * value(exp._args_[1])
        else:
            denom = 1.0 * exp._args_[1]
    else:
        res = _collect_standard_repn(
            exp._args_[1], 1, idMap, compute_values, verbose, quadratic
        )
        if (
            not (res.nonl.__class__ in native_numeric_types and res.nonl == 0)
            or len(res.linear) > 0
            or (quadratic and len(res.quadratic) > 0)
        ):
            # Denominator is variable, give up: this is nonlinear
            return Results(nonl=multiplier * exp)
        else:
            # Denominaor ended up evaluating to a constant
            denom = 1.0 * res.constant
    if denom.__class__ in native_numeric_types and denom == 0:
        raise ZeroDivisionError

    if (
        exp._args_[0].__class__ in native_numeric_types
        or not exp._args_[0].is_potentially_variable()
    ):
        num = exp._args_[0]
        if compute_values:
            num = value(num)
        return Results(constant=multiplier * num / denom)

    return _collect_standard_repn(
        exp._args_[0], multiplier / denom, idMap, compute_values, verbose, quadratic
    )


def _collect_branching_expr(exp, multiplier, idMap, compute_values, verbose, quadratic):
    _if, _then, _else = exp.args
    if _if.__class__ in native_types:
        if_val = _if
    elif not _if.is_potentially_variable():
        if compute_values:
            if_val = value(_if)
        else:
            return Results(nonl=multiplier * exp)
    else:
        res = _collect_standard_repn(_if, 1, idMap, compute_values, verbose, quadratic)
        if (
            not (res.nonl.__class__ in native_numeric_types and res.nonl == 0)
            or len(res.linear) > 0
            or (quadratic and len(res.quadratic) > 0)
        ):
            return Results(nonl=multiplier * exp)
        elif res.constant.__class__ in native_numeric_types:
            if_val = res.constant
        else:
            return Results(constant=multiplier * exp)
    if if_val:
        if _then.__class__ in native_numeric_types:
            return Results(constant=multiplier * _then)
        return _collect_standard_repn(
            _then, multiplier, idMap, compute_values, verbose, quadratic
        )
    else:
        if _else.__class__ in native_numeric_types:
            return Results(constant=multiplier * _else)
        return _collect_standard_repn(
            _else, multiplier, idMap, compute_values, verbose, quadratic
        )


def _collect_nonl(exp, multiplier, idMap, compute_values, verbose, quadratic):
    res = _collect_standard_repn(
        exp._args_[0], 1, idMap, compute_values, verbose, quadratic
    )
    if (
        not (res.nonl.__class__ in native_numeric_types and res.nonl == 0)
        or len(res.linear) > 0
        or (quadratic and len(res.quadratic) > 0)
    ):
        return Results(nonl=multiplier * exp)
    if compute_values:
        return Results(constant=multiplier * exp._apply_operation([res.constant]))
    else:
        return Results(constant=multiplier * exp)


def _collect_negation(exp, multiplier, idMap, compute_values, verbose, quadratic):
    return _collect_standard_repn(
        exp._args_[0], -1 * multiplier, idMap, compute_values, verbose, quadratic
    )


#
# TODO - Verify if code is used
#
def _collect_const(exp, multiplier, idMap, compute_values, verbose, quadratic):
    if compute_values:
        return Results(constant=multiplier * value(exp))
    else:
        return Results(constant=multiplier * exp)


def _collect_identity(exp, multiplier, idMap, compute_values, verbose, quadratic):
    if exp._args_[0].__class__ in native_numeric_types:
        return Results(constant=multiplier * exp._args_[0])
    if not exp._args_[0].is_potentially_variable():
        if compute_values:
            return Results(constant=multiplier * value(exp._args_[0]))
        else:
            return Results(constant=multiplier * exp._args_[0])
    return _collect_standard_repn(
        exp.expr, multiplier, idMap, compute_values, verbose, quadratic
    )


def _collect_linear(exp, multiplier, idMap, compute_values, verbose, quadratic):
    ans = Results()
    if compute_values:
        ans.constant = multiplier * value(exp.constant)
    else:
        ans.constant = multiplier * exp.constant

    linear = {}
    linear_vars = {}
    for c, v in zip(exp.linear_coefs, exp.linear_vars):
        if v.fixed:
            if compute_values:
                ans.constant += multiplier * value(c) * value(v)
            else:
                ans.constant += multiplier * c * v
        else:
            key = id(v)
            if compute_values:
                if key in linear:
                    linear[key] += multiplier * value(c)
                else:
                    linear[key] = multiplier * value(c)
                    linear_vars[key] = v
            else:
                if key in linear:
                    linear[key] += multiplier * c
                else:
                    linear[key] = multiplier * c
                    linear_vars[key] = v
    for id_, coef in linear.items():
        if coef.__class__ in native_numeric_types and not coef:
            continue
        if id_ in idMap[None]:
            key = idMap[None][id_]
        else:
            key = len(idMap) - 1
            idMap[None][id_] = key
            idMap[key] = linear_vars[id_]
        ans.linear[key] = coef
    return ans


def _collect_comparison(exp, multiplier, idMap, compute_values, verbose, quadratic):
    if multiplier != 1:
        # this *will* generate an exception with the new relational expressions
        exp = multiplier * exp
    return Results(nonl=exp)


def _collect_external_fn(exp, multiplier, idMap, compute_values, verbose, quadratic):
    if compute_values and exp.is_fixed():
        return Results(constant=multiplier * value(exp))
    return Results(nonl=multiplier * exp)


_repn_collectors = {
    EXPR.SumExpression: _collect_sum,
    EXPR.ProductExpression: _collect_prod,
    EXPR.MonomialTermExpression: _collect_term,
    EXPR.PowExpression: _collect_pow,
    EXPR.DivisionExpression: _collect_division,
    EXPR.Expr_ifExpression: _collect_branching_expr,
    EXPR.UnaryFunctionExpression: _collect_nonl,
    EXPR.AbsExpression: _collect_nonl,
    EXPR.NegationExpression: _collect_negation,
    EXPR.LinearExpression: _collect_linear,
    EXPR.InequalityExpression: _collect_comparison,
    EXPR.RangedExpression: _collect_comparison,
    EXPR.EqualityExpression: _collect_comparison,
    EXPR.ExternalFunctionExpression: _collect_external_fn,
    # ConnectorData          : _collect_linear_connector,
    # ScalarConnector         : _collect_linear_connector,
    ParamData: _collect_const,
    ScalarParam: _collect_const,
    # param.Param             : _collect_linear_const,
    # parameter               : _collect_linear_const,
    NumericConstant: _collect_const,
    VarData: _collect_var,
    ScalarVar: _collect_var,
    Var: _collect_var,
    variable: _collect_var,
    IVariable: _collect_var,
    ExpressionData: _collect_identity,
    ScalarExpression: _collect_identity,
    expression: _collect_identity,
    noclone: _collect_identity,
    NamedExpressionData: _collect_identity,
    Expression: _collect_identity,
    ObjectiveData: _collect_identity,
    ScalarObjective: _collect_identity,
    objective: _collect_identity,
}


def _collect_standard_repn(exp, multiplier, idMap, compute_values, verbose, quadratic):
    fn = _repn_collectors.get(exp.__class__, None)
    if fn is not None:
        return fn(exp, multiplier, idMap, compute_values, verbose, quadratic)
    #
    # Catch any known numeric constants
    #
    if exp.__class__ in native_numeric_types or not exp.is_potentially_variable():
        return _collect_const(
            exp, multiplier, idMap, compute_values, verbose, quadratic
        )
    #
    # These are types that might be extended using duck typing.
    #
    try:
        if exp.is_variable_type():
            fn = _collect_var
        if exp.is_named_expression_type():
            fn = _collect_identity
    except AttributeError:  # TODO: coverage?
        pass
    if fn is not None:
        _repn_collectors[exp.__class__] = fn
        return fn(exp, multiplier, idMap, compute_values, verbose, quadratic)
    raise ValueError(
        "Unexpected expression (type %s)" % type(exp).__name__
    )  # TODO: coverage?


def _generate_standard_repn(
    expr, idMap=None, compute_values=True, verbose=False, quadratic=True, repn=None
):
    if expr.__class__ is EXPR.SumExpression:
        #
        # This is the common case, so start collecting the sum
        #
        ans = _collect_sum(expr, 1, idMap, compute_values, verbose, quadratic)
    else:
        #
        # Call generic recursive logic
        #
        ans = _collect_standard_repn(expr, 1, idMap, compute_values, verbose, quadratic)
    #
    # Create the final object here from 'ans'
    #
    repn.constant = ans.constant
    #
    # Create a list (tuple) of the variables and coefficients
    #
    v = []
    c = []
    for key, val in ans.linear.items():
        if val.__class__ in native_numeric_types:
            if not val:
                continue
        elif val.is_constant():  # TODO: coverage?
            if value(val) == 0:
                continue
        v.append(idMap[key])
        c.append(val)
    repn.linear_vars = tuple(v)
    repn.linear_coefs = tuple(c)

    if quadratic:
        repn.quadratic_vars = []
        repn.quadratic_coefs = []
        for key in ans.quadratic:
            val = ans.quadratic[key]
            if val.__class__ in native_numeric_types:
                if val == 0:  # TODO: coverage?
                    continue
            elif val.is_constant():  # TODO: coverage?
                if value(val) == 0:
                    continue
            repn.quadratic_vars.append((idMap[key[0]], idMap[key[1]]))
            repn.quadratic_coefs.append(val)
        repn.quadratic_vars = tuple(repn.quadratic_vars)
        repn.quadratic_coefs = tuple(repn.quadratic_coefs)
        v = []
        c = []
        for key in ans.quadratic:
            v.append((idMap[key[0]], idMap[key[1]]))
            c.append(ans.quadratic[key])
        repn.quadratic_vars = tuple(v)
        repn.quadratic_coefs = tuple(c)

    if ans.nonl is not None and not isclose_const(ans.nonl, 0):
        repn.nonlinear_expr = ans.nonl
        repn.nonlinear_vars = []
        for v_ in EXPR.identify_variables(repn.nonlinear_expr, include_fixed=False):
            repn.nonlinear_vars.append(v_)
            #
            # Update idMap in case we skipped nonlinear sub-expressions
            #
            # Q: Should we skip nonlinear sub-expressions?
            #
            id_ = id(v_)
            if id_ in idMap[None]:
                key = idMap[None][id_]
            else:
                key = len(idMap) - 1
                idMap[None][id_] = key
                idMap[key] = v_
        repn.nonlinear_vars = tuple(repn.nonlinear_vars)

    return repn


"""
WEH - This code assumes the expression is linear and fills in a dictionary.
      This avoids creating temporary Results objects, but in practice that's
      not a big win.  Hence, this is deprecated.

##-----------------------------------------------------------------------
##
## Logic for _generate_linear_standard_repn
##
##-----------------------------------------------------------------------

def _linear_collect_sum(exp, multiplier, idMap, compute_values, verbose, coef):
    varkeys = idMap[None]

    for e_ in itertools.islice(exp._args_, exp.nargs()):
        if e_.__class__ in native_numeric_types:
            coef[None] += multiplier*e_

        elif e_.is_variable_type():
            if e_.fixed:
                if compute_values:
                    coef[None] += multiplier*e_.value
                else:
                    coef[None] += multiplier*e_
            else:
                id_ = id(e_)
                if id_ in varkeys:
                    key = varkeys[id_]
                else:
                    key = len(idMap) - 1
                    varkeys[id_] = key
                    idMap[key] = e_
                if key in coef:
                    coef[key] += multiplier
                else:
                    coef[key] = multiplier

        elif not e_.is_potentially_variable():
            if compute_values:
                coef[None] += multiplier * value(e_)
            else:
                coef[None] += multiplier * e_

        elif e_.__class__ is EXPR.NegationExpression:
            arg = e_._args_[0]
            if arg.is_variable_type():
                if arg.fixed:
                    if compute_values:
                        coef[None] -= multiplier*arg.value
                    else:
                        coef[None] -= multiplier*arg
                else:
                    id_ = id(arg)
                    if id_ in varkeys:
                        key = varkeys[id_]
                    else:
                        key = len(idMap) - 1
                        varkeys[id_] = key
                        idMap[key] = arg
                    if key in coef:
                        coef[key] -= multiplier
                    else:
                        coef[key] = -1 * multiplier
            else:
                _collect_linear_standard_repn(arg, -1*multiplier, idMap, compute_values, verbose, coef)

        elif e_.__class__ is EXPR.MonomialTermExpression:
            if compute_values:
                lhs = value(e_._args_[0])
            else:
                lhs = e_._args_[0]
            if e_._args_[1].fixed:
                if compute_values:
                    coef[None] += multiplier*lhs*value(e_._args_[1])
                else:
                    coef[None] += multiplier*lhs*e_._args_[1]
            else:
                id_ = id(e_._args_[1])
                if id_ in varkeys:
                    key = varkeys[id_]
                else:
                    key = len(idMap) - 1
                    varkeys[id_] = key
                    idMap[key] = e_._args_[1]
                if key in coef:
                    coef[key] += multiplier*lhs
                else:
                    coef[key] = multiplier*lhs

        else:
            _collect_linear_standard_repn(e_, multiplier, idMap, compute_values, verbose, coef)

def _linear_collect_linear(exp, multiplier, idMap, compute_values, verbose, coef):
    varkeys = idMap[None]

    if compute_values:
        coef[None] += multiplier*value(exp.constant)
    else:
        coef[None] += multiplier*exp.constant

    for c,v in zip(exp.linear_coefs, exp.linear_vars):
        if v.fixed:
            if compute_values:
                coef[None] += multiplier*v.value
            else:
                coef[None] += multiplier*v
        else:
            id_ = id(v)
            if id_ in varkeys:
                key = varkeys[id_]
            else:
                key = len(idMap) - 1
                varkeys[id_] = key
                idMap[key] = v
            if compute_values:
                if key in coef:
                    coef[key] += multiplier*value(c)
                else:
                    coef[key] = multiplier*value(c)
            else:
                if key in coef:
                    coef[key] += multiplier*c
                else:
                    coef[key] = multiplier*c

def _linear_collect_term(exp, multiplier, idMap, compute_values, verbose, coef):
    #
    # LHS is a numeric value
    #
    if exp._args_[0].__class__ in native_numeric_types:
        if isclose_default(exp._args_[0],0):
            return
        _collect_linear_standard_repn(exp._args_[1], multiplier * exp._args_[0], idMap,
                                  compute_values, verbose, coef)
    #
    # LHS is a non-variable expression
    #
    else:
        if compute_values:
            val = value(exp._args_[0])
            if isclose_default(val,0):
                return
            _collect_linear_standard_repn(exp._args_[1], multiplier * val, idMap,
                                  compute_values, verbose, coef)
        else:
            _collect_linear_standard_repn(exp._args_[1], multiplier*exp._args_[0], idMap,
                                  compute_values, verbose, coef)

def _linear_collect_prod(exp, multiplier, idMap, compute_values, verbose, coef):
    #
    # LHS is a numeric value
    #
    if exp._args_[0].__class__ in native_numeric_types:
        if isclose_default(exp._args_[0],0):
            return
        _collect_linear_standard_repn(exp._args_[1], multiplier * exp._args_[0], idMap,
                                  compute_values, verbose, coef)
    #
    # LHS is a non-variable expression
    #
    elif not exp._args_[0].is_potentially_variable():
        if compute_values:
            val = value(exp._args_[0])
            if isclose_default(val,0):
                return
            _collect_linear_standard_repn(exp._args_[1], multiplier * val, idMap,
                                  compute_values, verbose, coef)
        else:
            _collect_linear_standard_repn(exp._args_[1], multiplier*exp._args_[0], idMap,
                                  compute_values, verbose, coef)
    #
    # The LHS should never be variable
    #
    elif exp._args_[0].is_fixed():
        if compute_values:
            val = value(exp._args_[0])
            if isclose_default(val,0):
                return
            _collect_linear_standard_repn(exp._args_[1], multiplier * val, idMap,
                                  compute_values, verbose, coef)
        else:
            _collect_linear_standard_repn(exp._args_[1], multiplier*exp._args_[0], idMap,
                                  compute_values, verbose, coef)
    else:
        if compute_values:
            val = value(exp._args_[1])
            if isclose_default(val,0):
                return
            _collect_linear_standard_repn(exp._args_[0], multiplier * val, idMap,
                                  compute_values, verbose, coef)
        else:
            _collect_linear_standard_repn(exp._args_[0], multiplier*exp._args_[1], idMap,
                                  compute_values, verbose, coef)

def _linear_collect_var(exp, multiplier, idMap, compute_values, verbose, coef):
    if exp.fixed:
        if compute_values:
            coef[None] += multiplier*value(exp)
        else:
            coef[None] += multiplier*exp
    else:
        id_ = id(exp)
        if id_ in idMap[None]:
            key = idMap[None][id_]
        else:
            key = len(idMap) - 1
            idMap[None][id_] = key
            idMap[key] = exp
        if key in coef:
            coef[key] += multiplier
        else:
            coef[key] = multiplier

def _linear_collect_negation(exp, multiplier, idMap, compute_values, verbose, coef):
    _collect_linear_standard_repn(exp._args_[0], -1*multiplier, idMap, compute_values, verbose, coef)

def _linear_collect_identity(exp, multiplier, idMap, compute_values, verbose, coef):
    arg = exp._args_[0]
    if arg.__class__ in native_numeric_types:
        coef[None] += arg
    elif not arg.is_potentially_variable():
        if compute_values:
            coef[None] += value(arg)
        else:
            coef[None] += arg
    else:
        _collect_linear_standard_repn(exp.expr, multiplier, idMap, compute_values, verbose, coef)

def _linear_collect_branching_expr(exp, multiplier, idMap, compute_values, verbose, coef):
    if exp._if.__class__ in native_numeric_types:
        if_val = exp._if
    else:
        # If this value is not constant, then the expression is nonlinear.
        if_val = value(exp._if)
    if if_val:
        _collect_linear_standard_repn(exp._then, multiplier, idMap, compute_values, verbose, coef)
    else:
        _collect_linear_standard_repn(exp._else, multiplier, idMap, compute_values, verbose, coef)

def _linear_collect_pow(exp, multiplier, idMap, compute_values, verbose, coef):
    if exp._args_[1].__class__ in native_numeric_types:
        exponent = exp._args_[1]
    else:
        # If this value is not constant, then the expression is nonlinear.
        exponent = value(exp._args_[1])

    if exponent == 0:
        coef[None] += multiplier
    else: #exponent == 1
        _collect_linear_standard_repn(exp._args_[0], multiplier, idMap, compute_values, verbose, coef)


_linear_repn_collectors = {
    EXPR.SumExpression                          : _linear_collect_sum,
    EXPR.ProductExpression                      : _linear_collect_prod,
    EXPR.MonomialTermExpression                 : _linear_collect_term,
    EXPR.PowExpression                          : _linear_collect_pow,
    #EXPR.DivisionExpression                     : _linear_collect_division,
    EXPR.Expr_ifExpression                      : _linear_collect_branching_expr,
    #EXPR.UnaryFunctionExpression                : _linear_collect_nonl,
    #EXPR.AbsExpression                          : _linear_collect_nonl,
    EXPR.NegationExpression                     : _linear_collect_negation,
    EXPR.LinearExpression                       : _linear_collect_linear,
    #EXPR.InequalityExpression                   : _linear_collect_comparison,
    #EXPR.RangedExpression                       : _linear_collect_comparison,
    #EXPR.EqualityExpression                     : _linear_collect_comparison,
    #EXPR.ExternalFunctionExpression             : _linear_collect_external_fn,
    ##EXPR.LinearSumExpression               : _collect_linear_sum,
    ##ConnectorData          : _collect_linear_connector,
    ##ScalarConnector         : _collect_linear_connector,
    ##param.ParamData        : _collect_linear_const,
    ##param.ScalarParam       : _collect_linear_const,
    ##param.Param             : _collect_linear_const,
    ##parameter               : _collect_linear_const,
    VarData                             : _linear_collect_var,
    ScalarVar                                   : _linear_collect_var,
    Var                                         : _linear_collect_var,
    variable                                    : _linear_collect_var,
    IVariable                                   : _linear_collect_var,
    ExpressionData                      : _linear_collect_identity,
    ScalarExpression                            : _linear_collect_identity,
    expression                                  : _linear_collect_identity,
    noclone                                     : _linear_collect_identity,
    NamedExpressionData                             : _linear_collect_identity,
    Expression                                  : _linear_collect_identity,
    ObjectiveData                       : _linear_collect_identity,
    ScalarObjective                             : _linear_collect_identity,
    objective                                   : _linear_collect_identity,
    }


def _collect_linear_standard_repn(exp, multiplier, idMap, compute_values, verbose, coefs):
    fn = _linear_repn_collectors.get(exp.__class__, None)
    if fn is not None:
        return fn(exp, multiplier, idMap, compute_values, verbose, coefs)
    #
    # These are types that might be extended using duck typing.
    #
    try:
        if exp.is_variable_type():
            fn = _linear_collect_var
        if exp.is_named_expression_type():
            fn = _linear_collect_identity
    except AttributeError:
        pass
    if fn is not None:
        return fn(exp, multiplier, idMap, compute_values, verbose, coefs)
    raise ValueError( "Unexpected expression (type %s)" % type(exp).__name__)

def _generate_linear_standard_repn(expr, idMap=None, compute_values=True, verbose=False, repn=None):
    coef = {None:0}
    #
    # Call recursive logic
    #
    ans = _collect_linear_standard_repn(expr, 1, idMap, compute_values, verbose, coef)
    #
    # Create the final object here from 'ans'
    #
    repn.constant = coef[None]
    del coef[None]
    #
    # Create a list (tuple) of the variables and coefficients
    #
    # If we compute the values of constants, then we can skip terms with zero
    # coefficients
    #
    if compute_values:
        keys = list(key for key in coef if not isclose(coef[key],0))
    else:
        keys = list(coef.keys())
    repn.linear_vars = tuple(idMap[key] for key in keys)
    repn.linear_coefs = tuple(coef[key] for key in keys)

    return repn
"""


##-----------------------------------------------------------------------
##
## Functions to preprocess blocks
##
##-----------------------------------------------------------------------


def preprocess_block_objectives(block, idMap=None):
    # Get/Create the ComponentMap for the repn
    if not hasattr(block, '_repn'):
        block._repn = ComponentMap()
    block_repn = block._repn

    for objective_data in block.component_data_objects(
        Objective, active=True, descend_into=False
    ):
        if objective_data.expr is None:
            raise ValueError(
                "No expression has been defined for objective %s"
                % (objective_data.name)
            )

        try:
            repn = generate_standard_repn(objective_data.expr, idMap=idMap)
        except Exception:
            err = sys.exc_info()[1]
            logging.getLogger('pyomo.core').error(
                "exception generating a standard representation for objective %s: %s"
                % (objective_data.name, str(err))
            )
            raise

        block_repn[objective_data] = repn


def preprocess_block_constraints(block, idMap=None):
    # Get/Create the ComponentMap for the repn
    if not hasattr(block, '_repn'):
        block._repn = ComponentMap()
    block_repn = block._repn

    for constraint in block.component_objects(
        Constraint, active=True, descend_into=False
    ):
        preprocess_constraint(block, constraint, idMap=idMap, block_repn=block_repn)


def preprocess_constraint(block, constraint, idMap=None, block_repn=None):
    from pyomo.repn.beta.matrix import MatrixConstraint

    if isinstance(constraint, MatrixConstraint):
        return

    # Get/Create the ComponentMap for the repn
    if not hasattr(block, '_repn'):
        block._repn = ComponentMap()
    block_repn = block._repn

    for index, constraint_data in constraint.items():
        if not constraint_data.active:
            continue

        if constraint_data.body is None:
            raise ValueError(
                "No expression has been defined for the body "
                "of constraint %s" % (constraint_data.name)
            )

        try:
            repn = generate_standard_repn(constraint_data.body, idMap=idMap)
        except Exception:
            err = sys.exc_info()[1]
            logging.getLogger('pyomo.core').error(
                "exception generating a standard representation for "
                "constraint %s: %s" % (constraint_data.name, str(err))
            )
            raise

        block_repn[constraint_data] = repn


def preprocess_constraint_data(block, constraint_data, idMap=None, block_repn=None):
    # Get/Create the ComponentMap for the repn
    if not hasattr(block, '_repn'):
        block._repn = ComponentMap()
    block_repn = block._repn

    if constraint_data.body is None:
        raise ValueError(
            "No expression has been defined for the body "
            "of constraint %s" % (constraint_data.name)
        )

    try:
        repn = generate_standard_repn(constraint_data.body, idMap=idMap)
    except Exception:
        err = sys.exc_info()[1]
        logging.getLogger('pyomo.core').error(
            "exception generating a standard representation for "
            "constraint %s: %s" % (constraint_data.name, str(err))
        )
        raise

    block_repn[constraint_data] = repn



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""
This file contains a library of functions needed to construct
piecewise constraints for a Pyomo model. All piecewise types except
for SOS2, BIGM_SOS1, BIGM_BIN were taken from the paper:

Mixed-Integer Models for Non-separable Piecewise Linear Optimization:
Unifying framework and Extensions (Vielma, Nemhauser 2008).

TODO: Add regression tests for the following completed tasks
  - user not providing floats can be an major issue for BIGM's and MC
  - nonconvex/nonconcave functions - BIGM_SOS1, BIGM_SOS2 ***** possible edge case bug

Possible Extensions
  - Consider another piecewise rep ("SOS2_MANUAL"?) where we manually implement
    extra constraints to define an SOS2 set, this would be compatible with GLPK,
    http://winglpk.sourceforge.net/media/glpk-sos2_02.pdf
  - double check that LOG and DLOG reps really do require (2^n)+1 points, or can
    we just add integer cuts (or something more intelligent) in order to handle
    piecewise functions without 2^n polytopes
  - piecewise for functions of the form y = f(x1,x2,...)

"""

import logging
import math
import itertools
import operator
import types
import enum

from pyomo.common.log import is_debug_set
from pyomo.common.deprecation import RenamedClass, deprecation_warning
from pyomo.common.numeric_types import value
from pyomo.common.timing import ConstructionTimer
from pyomo.core.base.block import Block, BlockData
from pyomo.core.base.component import ModelComponentFactory
from pyomo.core.base.constraint import Constraint, ConstraintList
from pyomo.core.base.sos import SOSConstraint
from pyomo.core.base.var import Var, VarData, IndexedVar
from pyomo.core.base.set_types import PositiveReals, NonNegativeReals, Binary
from pyomo.core.base.util import flatten_tuple

logger = logging.getLogger('pyomo.core')


class PWRepn(str, enum.Enum):
    SOS2 = 'SOS2'
    BIGM_BIN = 'BIGM_BIN'
    BIGM_SOS1 = 'BIGM_SOS1'
    CC = 'CC'
    DCC = 'DCC'
    DLOG = 'DLOG'
    LOG = 'LOG'
    MC = 'MC'
    INC = 'INC'


class Bound(str, enum.Enum):
    Lower = 'Lower'
    Upper = 'Upper'
    Equal = 'Equal'


# BE SURE TO CHANGE THE PIECEWISE DOCSTRING
# IF THIS GETS CHANGED
_WARNING_TOLERANCE = 1e-8


def _isNonDecreasing(vals):
    """
    checks that list of points is
    nondecreasing
    """
    it = iter(vals)
    next(it)
    op = operator.ge
    return all(itertools.starmap(op, zip(it, vals)))


def _isNonIncreasing(vals):
    """
    checks that list of points is
    nonincreasing
    """
    it = iter(vals)
    next(it)
    op = operator.le
    return all(itertools.starmap(op, zip(it, vals)))


def _isPowerOfTwo(x):
    """
    checks that a number is a nonzero and positive power of 2
    """
    if x <= 0:
        return False
    else:
        return (x & (x - 1)) == 0


def _GrayCode(nbits):
    """
    Generates a GrayCode of nbits represented
    by a list of lists
    """
    bitset = [0 for i in range(nbits)]
    # important that we copy bitset each time
    graycode = [list(bitset)]

    for i in range(2, (1 << nbits) + 1):
        if i % 2:
            for j in range(-1, -nbits, -1):
                if bitset[j]:
                    bitset[j - 1] = bitset[j - 1] ^ 1
                    break
        else:
            bitset[-1] = bitset[-1] ^ 1
        # important that we copy bitset each time
        graycode.append(list(bitset))

    return graycode


def _characterize_function(name, tol, f_rule, model, points, *index):
    """
    Generates a list of range values and checks
    for convexity/concavity. Assumes domain points
    are sorted in increasing order.
    """
    # Make sure the list is a list of raw
    # numbers and not Pyomo Params or Expressions.
    # Failing to do this can generate strange
    # expression generation errors in the checks below
    points = [value(_p) for _p in points]

    if isinstance(f_rule, types.FunctionType):
        values = [f_rule(model, *flatten_tuple((index, x))) for x in points]
    elif f_rule.__class__ is dict:
        if len(index) == 1:
            values = f_rule[index[0]]
        else:
            values = f_rule[index]
    else:  # a list or tuple
        values = f_rule
    # Make sure the list is a list of raw
    # numbers and not Pyomo Params or Expressions.
    # Failing to do this can generate strange
    # expression generation errors in the checks below
    values = [value(_p) for _p in values]

    step = False
    try:
        slopes = [
            (values[i] - values[i - 1]) / (points[i] - points[i - 1])
            for i in range(1, len(points))
        ]
    except ZeroDivisionError:
        # we have a step function
        step = True
        slopes = [
            (
                (None)
                if (points[i] == points[i - 1])
                else ((values[i] - values[i - 1]) / (points[i] - points[i - 1]))
            )
            for i in range(1, len(points))
        ]

    # TODO: Warn when the slopes of two consecutive line
    #       segments are nearly equal since this is likely
    #       due to a user mistake and may cause issue with
    #       the solver.
    #       *** This is already done below but there
    #           is probably a more correct way
    #           to send this warning through Pyomo
    if not all(
        itertools.starmap(
            lambda x1, x2: (
                (True) if ((x1 is None) or (x2 is None)) else (abs(x1 - x2) > tol)
            ),
            zip(slopes, itertools.islice(slopes, 1, None)),
        )
    ):
        msg = (
            "**WARNING: Piecewise component '%s[%s]' has detected slopes of consecutive piecewise "
            "segments to be within "
            + str(tol)
            + " of one another. Refer to the Piecewise help "
            "documentation for information on how to disable this warning."
        )
        if index == ():
            index = None
        print(msg % (name, flatten_tuple(index)))

    if step is True:
        return 0, values, True
    if _isNonDecreasing(slopes):
        # convex
        return 1, values, False
    if _isNonIncreasing(slopes):
        # concave
        return -1, values, False
    return 0, values, False


class PiecewiseData(BlockData):
    """
    This class defines the base class for all linearization
    and piecewise constraint generators..
    """

    def __init__(self, parent):
        BlockData.__init__(self, parent)
        self._constructed = True
        self._bound_type = None
        self._domain_pts = None
        self._range_pts = None
        self._x = None
        self._y = None

    def updateBoundType(self, bound_type):
        self._bound_type = bound_type

    def updatePoints(self, domain_pts, range_pts):
        # ***Note: most (if not all) piecewise constraint generators
        #          assume the list of domain points is sorted.
        if not _isNonDecreasing(domain_pts):
            msg = "'%s' does not have a list of domain points that is non-decreasing"
            raise ValueError(msg % (self.name,))
        self._domain_pts = domain_pts
        self._range_pts = range_pts

    def build_constraints(self, functor, x_var, y_var):
        functor.construct(self, x_var, y_var)
        self.__dict__['_x'] = x_var
        self.__dict__['_y'] = y_var

    def referenced_variables(self):
        return (self._x, self._y)

    def __call__(self, x):
        if self._constructed is False:
            raise ValueError(
                "Piecewise component %s has not been constructed yet" % self.name
            )

        for i in range(len(self._domain_pts) - 1):
            xL = self._domain_pts[i]
            xU = self._domain_pts[i + 1]
            if (xL <= x) and (x <= xU):
                yL = self._range_pts[i]
                yU = self._range_pts[i + 1]
                if xL == xU:  # a step function
                    return yU
                return yL + ((yU - yL) / (xU - xL)) * (x - xL)
        raise ValueError(
            "The point %s is outside the list of domain "
            "points for Piecewise component %s. The valid "
            "point range is [%s,%s]."
            % (x, self.name, min(self._domain_pts), max(self._domain_pts))
        )


class _PiecewiseData(metaclass=RenamedClass):
    __renamed__new_class__ = PiecewiseData
    __renamed__version__ = '6.7.2'


class _SimpleSinglePiecewise:
    """
    Called when the piecewise points list has only two points
    """

    def construct(self, pblock, x_var, y_var):
        x_pts = pblock._domain_pts
        y_pts = pblock._range_pts
        bound_type = pblock._bound_type
        if None in [x_pts, y_pts, bound_type]:
            raise RuntimeError(
                "_SimpleSinglePiecewise: construct() called duringinvalid state."
            )

        # create a single linear constraint
        LHS = y_var
        F_AT_XO = y_pts[0]
        dF_AT_XO = (y_pts[1] - y_pts[0]) / (x_pts[1] - x_pts[0])
        X_MINUS_XO = x_var - x_pts[0]
        if bound_type == Bound.Upper:
            expr = LHS <= F_AT_XO + dF_AT_XO * X_MINUS_XO
        elif bound_type == Bound.Lower:
            expr = LHS >= F_AT_XO + dF_AT_XO * X_MINUS_XO
        elif bound_type == Bound.Equal:
            expr = LHS == F_AT_XO + dF_AT_XO * X_MINUS_XO
        else:
            raise ValueError("Invalid Bound for _SimpleSinglePiecewise object")
        pblock.single_line_constraint = Constraint(expr=expr)

        # In order to enforce the same behavior as actual piecewise
        # constraints, we constrain the domain variable between the
        # outer domain pts. But in order to prevent filling the model
        # with unnecessary constraints, we only do this when absolutely
        # necessary.
        if not x_var.lb is None and x_var.lb < x_pts[0]:
            pblock.simplified_piecewise_domain_constraint_lower = Constraint(
                expr=x_pts[0] <= x_var
            )
        if not x_var.ub is None and x_var.ub > x_pts[1]:
            pblock.simplified_piecewise_domain_constraint_upper = Constraint(
                expr=x_var <= x_pts[-1]
            )


class _SimplifiedPiecewise:
    """
    Called when piecewise constraints are simplified due to a lower bounding
    convex function or an upper bounding concave function
    """

    def construct(self, pblock, x_var, y_var):
        x_pts = pblock._domain_pts
        y_pts = pblock._range_pts
        bound_type = pblock._bound_type
        if None in [x_pts, y_pts, bound_type]:
            raise RuntimeError(
                "_SimplifiedPiecewise: construct() called during invalid state."
            )
        len_x_pts = len(x_pts)

        conlist = pblock.simplified_piecewise_constraint = ConstraintList()
        for i in range(len_x_pts - 1):
            F_AT_XO = y_pts[i]
            dF_AT_XO = (y_pts[i + 1] - y_pts[i]) / (x_pts[i + 1] - x_pts[i])
            XO = x_pts[i]
            if bound_type == Bound.Upper:
                conlist.add((0, -y_var + F_AT_XO + dF_AT_XO * (x_var - XO), None))
            elif bound_type == Bound.Lower:
                conlist.add((None, -y_var + F_AT_XO + dF_AT_XO * (x_var - XO), 0))
            else:
                raise ValueError("Invalid Bound for _SimplifiedPiecewise object")

        # In order to enforce the same behavior as actual piecewise
        # constraints, we constrain the domain variable between the
        # outer domain pts. But in order to prevent filling the model
        # with unnecessary constraints, we only do this when absolutely
        # necessary.
        if not x_var.lb is None and x_var.lb < x_pts[0]:
            pblock.simplified_piecewise_domain_constraint_lower = Constraint(
                expr=x_pts[0] <= x_var
            )
        if not x_var.ub is None and x_var.ub > x_pts[-1]:
            pblock.simplified_piecewise_domain_constraint_upper = Constraint(
                expr=x_var <= x_pts[-1]
            )


class _SOS2Piecewise:
    """
    Called to generate Piecewise constraint using the SOS2 formulation
    """

    def construct(self, pblock, x_var, y_var):
        x_pts = pblock._domain_pts
        y_pts = pblock._range_pts
        bound_type = pblock._bound_type
        if None in [x_pts, y_pts, bound_type]:
            raise RuntimeError(
                "_SOS2Piecewise: construct() called during invalid state."
            )
        len_x_pts = len(x_pts)

        # create indexers
        sos2_index = range(len_x_pts)

        # create vars
        sos2_y = pblock.SOS2_y = Var(sos2_index, within=NonNegativeReals)

        # create piecewise constraints
        conlist = pblock.SOS2_constraint = ConstraintList()
        conlist.add((x_var - sum(sos2_y[i] * x_pts[i] for i in sos2_index), 0))

        LHS = y_var
        RHS = sum(sos2_y[i] * y_pts[i] for i in sos2_index)
        expr = None
        if bound_type == Bound.Upper:
            conlist.add((None, LHS - RHS, 0))
        elif bound_type == Bound.Lower:
            conlist.add((0, LHS - RHS, None))
        elif bound_type == Bound.Equal:
            conlist.add((LHS - RHS, 0))
        else:
            raise ValueError("Invalid Bound for _SOS2Piecewise object")
        conlist.add((sum(sos2_y[j] for j in sos2_index), 1))

        def SOS2_rule(model):
            return [sos2_y[i] for i in sos2_index]

        pblock.SOS2_sosconstraint = SOSConstraint(initialize=SOS2_rule, sos=2)


class _DCCPiecewise:
    """
    Called to generate Piecewise constraint using the DCC formulation
    """

    def construct(self, pblock, x_var, y_var):
        x_pts = pblock._domain_pts
        y_pts = pblock._range_pts
        bound_type = pblock._bound_type
        if None in [x_pts, y_pts, bound_type]:
            raise RuntimeError(
                "_DCCPiecewise: construct() called during invalid state."
            )
        len_x_pts = len(x_pts)

        # create indexers
        polytopes = range(1, len_x_pts)
        vertices = range(1, len_x_pts + 1)

        def polytope_verts(p):
            return range(p, p + 2)

        # create vars
        pblock.DCC_lambda = Var(polytopes, vertices, within=PositiveReals)
        lmda = pblock.DCC_lambda
        pblock.DCC_bin_y = Var(polytopes, within=Binary)
        bin_y = pblock.DCC_bin_y

        # create piecewise constraints
        pblock.DCC_constraint1 = Constraint(
            expr=x_var
            == sum(
                lmda[p, v] * x_pts[v - 1] for p in polytopes for v in polytope_verts(p)
            )
        )

        LHS = y_var
        RHS = sum(
            lmda[p, v] * y_pts[v - 1] for p in polytopes for v in polytope_verts(p)
        )
        expr = None
        if bound_type == Bound.Upper:
            expr = LHS <= RHS
        elif bound_type == Bound.Lower:
            expr = LHS >= RHS
        elif bound_type == Bound.Equal:
            expr = LHS == RHS
        else:
            raise ValueError("Invalid Bound for _DCCPiecewise object")
        pblock.DCC_constraint2 = Constraint(expr=expr)

        def con3_rule(model, p):
            return bin_y[p] == sum(lmda[p, v] for v in polytope_verts(p))

        pblock.DCC_constraint3 = Constraint(polytopes, rule=con3_rule)
        pblock.DCC_constraint4 = Constraint(expr=sum(bin_y[p] for p in polytopes) == 1)


class _DLOGPiecewise:
    """
    Called to generate Piecewise constraint using the DLOG formulation
    """

    def _Branching_Scheme(self, L):
        """
        Branching scheme for DLOG
        """
        MAX = 2**L
        mylists1 = {}
        for i in range(1, L + 1):
            mylists1[i] = []
            start = 1
            step = int(MAX / (2**i))
            while start < MAX:
                mylists1[i].extend([j for j in range(start, start + step)])
                start += 2 * step

        biglist = range(1, MAX + 1)
        mylists2 = {}
        for i in sorted(mylists1.keys()):
            mylists2[i] = []
            for j in biglist:
                if j not in mylists1[i]:
                    mylists2[i].append(j)
            mylists2[i] = sorted(mylists2[i])

        return mylists1, mylists2

    def construct(self, pblock, x_var, y_var):
        if not _isPowerOfTwo(len(pblock._domain_pts) - 1):
            msg = "'%s' does not have a list of domain points with length (2^n)+1"
            raise ValueError(msg % (pblock.name,))
        x_pts = pblock._domain_pts
        y_pts = pblock._range_pts
        bound_type = pblock._bound_type
        if None in [x_pts, y_pts, bound_type]:
            raise RuntimeError(
                "_DLOGPiecewise: construct() called during invalid state."
            )
        len_x_pts = len(x_pts)

        # create branching schemes
        L_i = int(math.log(len_x_pts - 1, 2))
        B_ZERO, B_ONE = self._Branching_Scheme(L_i)

        # create indexers
        polytopes = range(1, len_x_pts)
        vertices = range(1, len_x_pts + 1)
        bin_y_index = range(1, L_i + 1)

        def polytope_verts(p):
            return range(p, p + 2)

        # create vars
        pblock.DLOG_lambda = Var(polytopes, vertices, within=PositiveReals)
        lmda = pblock.DLOG_lambda
        pblock.DLOG_bin_y = Var(bin_y_index, within=Binary)
        bin_y = pblock.DLOG_bin_y
        # create piecewise constraints
        pblock.DLOG_constraint1 = Constraint(
            expr=x_var
            == sum(
                lmda[p, v] * x_pts[v - 1] for p in polytopes for v in polytope_verts(p)
            )
        )

        LHS = y_var
        RHS = sum(
            lmda[p, v] * y_pts[v - 1] for p in polytopes for v in polytope_verts(p)
        )
        expr = None
        if bound_type == Bound.Upper:
            expr = LHS <= RHS
        elif bound_type == Bound.Lower:
            expr = LHS >= RHS
        elif bound_type == Bound.Equal:
            expr = LHS == RHS
        else:
            raise ValueError("Invalid Bound for _DLOGPiecewise object")
        pblock.DLOG_constraint2 = Constraint(expr=expr)
        pblock.DLOG_constraint3 = Constraint(
            expr=sum(lmda[p, v] for p in polytopes for v in polytope_verts(p)) == 1
        )

        def con4_rule(model, l):
            return (
                sum(lmda[p, v] for p in B_ZERO[l] for v in polytope_verts(p))
                <= bin_y[l]
            )

        pblock.DLOG_constraint4 = Constraint(bin_y_index, rule=con4_rule)

        def con5_rule(model, l):
            return sum(lmda[p, v] for p in B_ONE[l] for v in polytope_verts(p)) <= (
                1 - bin_y[l]
            )

        pblock.DLOG_constraint5 = Constraint(bin_y_index, rule=con5_rule)


class _CCPiecewise:
    """
    Called to generate Piecewise constraint using the CC formulation
    """

    def construct(self, pblock, x_var, y_var):
        x_pts = pblock._domain_pts
        y_pts = pblock._range_pts
        bound_type = pblock._bound_type
        if None in [x_pts, y_pts, bound_type]:
            raise RuntimeError("_CCPiecewise: construct() called during invalid state.")
        len_x_pts = len(x_pts)

        # create indexers
        polytopes = range(1, len_x_pts)
        vertices = range(1, len_x_pts + 1)

        def vertex_polys(v):
            if v == 1:
                return [v]
            if v == len_x_pts:
                return [v - 1]
            else:
                return [v - 1, v]

        # create vars
        pblock.CC_lambda = Var(vertices, within=NonNegativeReals)
        lmda = pblock.CC_lambda
        pblock.CC_bin_y = Var(polytopes, within=Binary)
        bin_y = pblock.CC_bin_y
        # create piecewise constraints
        pblock.CC_constraint1 = Constraint(
            expr=x_var == sum(lmda[v] * x_pts[v - 1] for v in vertices)
        )

        LHS = y_var
        RHS = sum(lmda[v] * y_pts[v - 1] for v in vertices)
        expr = None
        if bound_type == Bound.Upper:
            expr = LHS <= RHS
        elif bound_type == Bound.Lower:
            expr = LHS >= RHS
        elif bound_type == Bound.Equal:
            expr = LHS == RHS
        else:
            raise ValueError("Invalid Bound for _CCPiecewise object")
        pblock.CC_constraint2 = Constraint(expr=expr)
        pblock.CC_constraint3 = Constraint(expr=sum(lmda[v] for v in vertices) == 1)

        def con4_rule(model, v):
            return lmda[v] <= sum(bin_y[p] for p in vertex_polys(v))

        pblock.CC_constraint4 = Constraint(vertices, rule=con4_rule)
        pblock.CC_constraint5 = Constraint(expr=sum(bin_y[p] for p in polytopes) == 1)


class _LOGPiecewise:
    """
    Called to generate Piecewise constraint using the LOG formulation
    """

    def _Branching_Scheme(self, n):
        """
        Branching scheme for LOG, requires a gray code
        """
        BIGL = 2**n
        S = range(1, n + 1)
        # turn the GrayCode into a dictionary indexed
        # starting at 1
        G = {k: v for k, v in enumerate(_GrayCode(n), start=1)}

        L = {
            s: [
                k + 1
                for k in range(BIGL + 1)
                if ((k == 0) or (G[k][s - 1] == 1))
                and ((k == BIGL) or (G[k + 1][s - 1] == 1))
            ]
            for s in S
        }
        R = {
            s: [
                k + 1
                for k in range(BIGL + 1)
                if ((k == 0) or (G[k][s - 1] == 0))
                and ((k == BIGL) or (G[k + 1][s - 1] == 0))
            ]
            for s in S
        }

        return S, L, R

    def construct(self, pblock, x_var, y_var):
        if not _isPowerOfTwo(len(pblock._domain_pts) - 1):
            msg = "'%s' does not have a list of domain points with length (2^n)+1"
            raise ValueError(msg % (pblock.name,))
        x_pts = pblock._domain_pts
        y_pts = pblock._range_pts
        bound_type = pblock._bound_type
        if None in [x_pts, y_pts, bound_type]:
            raise RuntimeError(
                "_LOGPiecewise: construct() called during invalid state."
            )
        len_x_pts = len(x_pts)

        # create branching schemes
        L_i = int(math.log(len_x_pts - 1, 2))
        S_i, B_LEFT, B_RIGHT = self._Branching_Scheme(L_i)

        # create indexers
        polytopes = range(1, len_x_pts)
        vertices = range(1, len_x_pts + 1)
        bin_y_index = S_i

        # create vars
        pblock.LOG_lambda = Var(vertices, within=NonNegativeReals)
        lmda = pblock.LOG_lambda
        pblock.LOG_bin_y = Var(bin_y_index, within=Binary)
        bin_y = pblock.LOG_bin_y
        # create piecewise constraints
        pblock.LOG_constraint1 = Constraint(
            expr=x_var == sum(lmda[v] * x_pts[v - 1] for v in vertices)
        )

        LHS = y_var
        RHS = sum(lmda[v] * y_pts[v - 1] for v in vertices)
        expr = None
        if bound_type == Bound.Upper:
            expr = LHS <= RHS
        elif bound_type == Bound.Lower:
            expr = LHS >= RHS
        elif bound_type == Bound.Equal:
            expr = LHS == RHS
        else:
            raise ValueError("Invalid Bound for _LOGPiecewise object")
        pblock.LOG_constraint2 = Constraint(expr=expr)
        pblock.LOG_constraint3 = Constraint(expr=sum(lmda[v] for v in vertices) == 1)

        def con4_rule(model, s):
            return sum(lmda[v] for v in B_LEFT[s]) <= bin_y[s]

        pblock.LOG_constraint4 = Constraint(bin_y_index, rule=con4_rule)

        def con5_rule(model, s):
            return sum(lmda[v] for v in B_RIGHT[s]) <= (1 - bin_y[s])

        pblock.LOG_constraint5 = Constraint(bin_y_index, rule=con5_rule)


class _MCPiecewise:
    """
    Called to generate Piecewise constraint using the MC formulation
    """

    def construct(self, pblock, x_var, y_var):
        x_pts = pblock._domain_pts
        y_pts = pblock._range_pts
        bound_type = pblock._bound_type
        if None in [x_pts, y_pts, bound_type]:
            raise RuntimeError("_MCPiecewise: construct() called during invalid state.")
        len_x_pts = len(x_pts)

        # create indexers
        polytopes = range(1, len_x_pts)

        # create constants
        SLOPE = {
            p: (y_pts[p] - y_pts[p - 1]) / (x_pts[p] - x_pts[p - 1]) for p in polytopes
        }
        INTERSEPT = {p: y_pts[p - 1] - (SLOPE[p] * x_pts[p - 1]) for p in polytopes}

        # create vars
        pblock.MC_poly_x = Var(polytopes)
        poly_x = pblock.MC_poly_x
        pblock.MC_bin_y = Var(polytopes, within=Binary)
        bin_y = pblock.MC_bin_y
        # create piecewise constraints
        pblock.MC_constraint1 = Constraint(
            expr=x_var == sum(poly_x[p] for p in polytopes)
        )

        LHS = y_var
        RHS = sum(poly_x[p] * SLOPE[p] + bin_y[p] * INTERSEPT[p] for p in polytopes)
        expr = None
        if bound_type == Bound.Upper:
            expr = LHS <= RHS
        elif bound_type == Bound.Lower:
            expr = LHS >= RHS
        elif bound_type == Bound.Equal:
            expr = LHS == RHS
        else:
            raise ValueError("Invalid Bound for _INCPiecewise object")
        pblock.MC_constraint2 = Constraint(expr=expr)

        def con3_rule(model, p):
            return bin_y[p] * x_pts[p - 1] <= poly_x[p]

        pblock.MC_constraint3 = Constraint(polytopes, rule=con3_rule)

        def con4_rule(model, p):
            return poly_x[p] <= bin_y[p] * x_pts[p]

        pblock.MC_constraint4 = Constraint(polytopes, rule=con4_rule)
        pblock.MC_constraint5 = Constraint(expr=sum(bin_y[p] for p in polytopes) == 1)


class _INCPiecewise:
    """
    Called to generate Piecewise constraint using the INC formulation
    """

    def construct(self, pblock, x_var, y_var):
        x_pts = pblock._domain_pts
        y_pts = pblock._range_pts
        bound_type = pblock._bound_type
        if None in [x_pts, y_pts, bound_type]:
            raise RuntimeError(
                "_INCPiecewise: construct() called during invalid state."
            )
        len_x_pts = len(x_pts)

        # create indexers
        polytopes = range(1, len_x_pts)
        bin_y_index = range(1, len_x_pts - 1)

        # create vars
        pblock.INC_delta = Var(polytopes)
        delta = pblock.INC_delta
        delta[1].setub(1)
        delta[len_x_pts - 1].setlb(0)
        pblock.INC_bin_y = Var(bin_y_index, within=Binary)
        bin_y = pblock.INC_bin_y
        # create piecewise constraints
        pblock.INC_constraint1 = Constraint(
            expr=x_var
            == x_pts[0] + sum(delta[p] * (x_pts[p] - x_pts[p - 1]) for p in polytopes)
        )

        LHS = y_var
        RHS = y_pts[0] + sum(delta[p] * (y_pts[p] - y_pts[p - 1]) for p in polytopes)
        expr = None
        if bound_type == Bound.Upper:
            expr = LHS <= RHS
        elif bound_type == Bound.Lower:
            expr = LHS >= RHS
        elif bound_type == Bound.Equal:
            expr = LHS == RHS
        else:
            raise ValueError("Invalid Bound for _INCPiecewise object")
        pblock.INC_constraint2 = Constraint(expr=expr)

        def con3_rule(model, p):
            if p != polytopes[-1]:
                return delta[p + 1] <= bin_y[p]
            else:
                return Constraint.Skip

        pblock.INC_constraint3 = Constraint(polytopes, rule=con3_rule)

        def con4_rule(model, p):
            if p != polytopes[-1]:
                return bin_y[p] <= delta[p]
            else:
                return Constraint.Skip

        pblock.INC_constraint4 = Constraint(polytopes, rule=con4_rule)


class _BIGMPiecewise:
    """
    Called to generate Piecewise constraint using the BIGM formulation
    """

    def __init__(self, binary=True):
        self.binary = binary
        if not (self.binary in [True, False]):
            raise ValueError(
                "_BIGMPiecewise must be initialized with the binary "
                "flag set to True or False (choose one)."
            )

    def construct(self, pblock, x_var, y_var):
        # The BIGM methods currently determine tightest possible M
        # values. This method is implemented in such a way that
        # binary/sos1 variables are not created when this M is zero
        tag = ""
        x_pts = pblock._domain_pts
        y_pts = pblock._range_pts
        bound_type = pblock._bound_type
        if None in [x_pts, y_pts, bound_type]:
            raise RuntimeError(
                "_BIGMPiecewise: construct() called during invalid state."
            )
        len_x_pts = len(x_pts)

        if self.binary is True:
            tag += "bin"
        else:
            tag += "sos1"

        # generate tightest bigM values
        OPT_M = {}
        OPT_M['UB'] = {}
        OPT_M['LB'] = {}

        if bound_type in [Bound.Upper, Bound.Equal]:
            OPT_M['UB'] = self._find_M(x_pts, y_pts, Bound.Upper)
        if bound_type in [Bound.Lower, Bound.Equal]:
            OPT_M['LB'] = self._find_M(x_pts, y_pts, Bound.Lower)

        all_keys = set(OPT_M['UB'].keys()).union(OPT_M['LB'].keys())
        full_indices = []
        full_indices.extend(range(1, len_x_pts))
        bigm_y_index = None
        bigm_y = None
        if len(all_keys) > 0:
            bigm_y_index = all_keys

            def y_domain():
                if self.binary is True:
                    return Binary
                else:
                    return NonNegativeReals

            setattr(pblock, tag + '_y', Var(bigm_y_index, within=y_domain()))
            bigm_y = getattr(pblock, tag + '_y')

        def con1_rule(model, i):
            if bound_type in [Bound.Upper, Bound.Equal]:
                rhs = 1.0
                if i not in OPT_M['UB']:
                    rhs *= 0.0
                else:
                    rhs *= OPT_M['UB'][i] * (1 - bigm_y[i])
                return (
                    y_var
                    - y_pts[i - 1]
                    - ((y_pts[i] - y_pts[i - 1]) / (x_pts[i] - x_pts[i - 1]))
                    * (x_var - x_pts[i - 1])
                    <= rhs
                )
            elif bound_type == Bound.Lower:
                rhs = 1.0
                if i not in OPT_M['LB']:
                    rhs *= 0.0
                else:
                    rhs *= OPT_M['LB'][i] * (1 - bigm_y[i])
                return (
                    y_var
                    - y_pts[i - 1]
                    - ((y_pts[i] - y_pts[i - 1]) / (x_pts[i] - x_pts[i - 1]))
                    * (x_var - x_pts[i - 1])
                    >= rhs
                )

        def con2_rule(model):
            expr = [bigm_y[i] for i in range(1, len_x_pts) if i in all_keys]
            if len(expr) > 0:
                return sum(expr) == 1
            else:
                return Constraint.Skip

        def conAFF_rule(model, i):
            rhs = 1.0
            if i not in OPT_M['LB']:
                rhs *= 0.0
            else:
                rhs *= OPT_M['LB'][i] * (1 - bigm_y[i])
            return (
                y_var
                - y_pts[i - 1]
                - ((y_pts[i] - y_pts[i - 1]) / (x_pts[i] - x_pts[i - 1]))
                * (x_var - x_pts[i - 1])
                >= rhs
            )

        pblock.BIGM_constraint1 = Constraint(full_indices, rule=con1_rule)
        if len(all_keys) > 0:
            pblock.BIGM_constraint2 = Constraint(rule=con2_rule)
        if bound_type == Bound.Equal:
            pblock.BIGM_constraint3 = Constraint(full_indices, rule=conAFF_rule)

        if len(all_keys) > 0:
            if self.binary is False:
                pblock.BIGM_constraint4 = SOSConstraint(var=bigm_y, sos=1)

        # In order to enforce the same behavior as actual piecewise
        # constraints, we constrain the domain variable between the
        # outer domain pts. But in order to prevent filling the model
        # with unnecessary constraints, we only do this when absolutely
        # necessary.
        if not x_var.lb is None and x_var.lb < x_pts[0]:
            pblock.bigm_domain_constraint_lower = Constraint(expr=x_pts[0] <= x_var)
        if not x_var.ub is None and x_var.ub > x_pts[-1]:
            pblock.bigm_domain_constraint_upper = Constraint(expr=x_var <= x_pts[-1])

    def _M_func(self, a, Fa, b, Fb, c, Fc):
        return Fa - Fb - ((a - b) * ((Fc - Fb) / (c - b)))

    def _find_M(self, x_pts, y_pts, bound_type):
        len_x_pts = len(x_pts)
        _self_M_func = self._M_func

        M_final = {}
        for j in range(1, len_x_pts):
            index = j
            if bound_type == Bound.Lower:
                M_final[index] = min(
                    [
                        0.0,
                        min(
                            [
                                _self_M_func(
                                    x_pts[k],
                                    y_pts[k],
                                    x_pts[j - 1],
                                    y_pts[j - 1],
                                    x_pts[j],
                                    y_pts[j],
                                )
                                for k in range(len_x_pts)
                            ]
                        ),
                    ]
                )
            elif bound_type == Bound.Upper:
                M_final[index] = max(
                    [
                        0.0,
                        max(
                            [
                                _self_M_func(
                                    x_pts[k],
                                    y_pts[k],
                                    x_pts[j - 1],
                                    y_pts[j - 1],
                                    x_pts[j],
                                    y_pts[j],
                                )
                                for k in range(len_x_pts)
                            ]
                        ),
                    ]
                )
            else:
                raise ValueError("Invalid Bound passed to _find_M function")
            if M_final[index] == 0.0:
                del M_final[index]
        return M_final


@ModelComponentFactory.register(
    "Constraints that contain piecewise linear expressions."
)
class Piecewise(Block):
    r"""Adds piecewise constraints to a Pyomo model for functions of the
    form, y = f(x).

    Examples
    --------

    .. code::

        model.const = Piecewise(index_1,...,index_n,yvar,xvar,**Keywords)
        model.const = Piecewise(yvar,xvar,**Keywords)

    Parameters
    ----------
    pw_pts : dict
        A dictionary of lists (keys are index set) or a single list (for
        the non-indexed case or when an identical set of breakpoints is
        used across all indices) defining the set of domain breakpoints
        for the piecewise linear function. **ALWAYS REQUIRED**

    pw_repn : str

        Indicates the type of piecewise representation to use. This can
        have a major impact on solver performance.  Choices: (Default
        'SOS2')

           - ``SOS2``: +
               Standard representation using sos2 constraints
           - ``BIGM_BIN``:
               BigM constraints with binary variables.  Theoretically
               tightest M values are automatically determined.
           - ``BIGM_SOS1``:
               BigM constraints with sos1 variables.  Theoretically
               tightest M values are automatically determined.
           - ``DCC``: \*+
               Disaggregated convex combination model
           - ``DLOG``: \*+
               Logarithmic disaggregated convex combination model
           - ``CC``: \*+
               Convex combination model
           - ``LOG``: \*+
               Logarithmic branching convex combination
           - ``MC``: \*
               Multiple choice model
           - ``INC``: \*+
               Incremental (delta) method

        .. note::

            \+\: Supports step functions

            \*\: From "Mixed-Integer Models for Non-separable Piecewise Linear
            Optimization: Unifying framework and Extensions" (Vielma,
            Nemhauser 2008)

        .. seealso::
            Refer to the optional 'force_pw' keyword.

    pw_constr_type : str
        Indicates the bound type of the piecewise function. Choices:

           - ``UB`` - y variable is bounded above by piecewise function
           - ``LB`` - y variable is bounded below by piecewise function
           - ``EQ`` - y variable is equal to the piecewise function

    f_rule : f(model,i,j,...,x), {}, [], ()
        An object that returns a numeric value that is the range value
        corresponding to each piecewise domain point. For functions, the
        first argument must be a Pyomo model. The last argument is the
        domain value at which the function evaluates (Not a Pyomo
        Var). Intermediate arguments are the corresponding indices of
        the Piecewise component (if any).  Otherwise, the object can be
        a dictionary of lists/tuples (with keys the same as the indexing
        set) or a singe list/tuple (when no indexing set is used or when
        all indices use an identical piecewise function).  Examples:

        .. code:: python

            # A function which changes with index
            def f(model,j,x):
                if (j == 2):
                    return x**2 + 1.0
                else:
                    return x**2 + 5.0

            # A nonlinear function
            f = lambda model, x: return exp(x) + value(model.p)
            # (where model.p is a Pyomo Param)

            # A step function
            f = [0,0,1,1,2,2]

    force_pw : bool
        Using the given function rule and pw_pts, a check for
        convexity/concavity is implemented. If (1) the function is
        convex and the piecewise constraints are lower bounds or if (2)
        the function is concave and the piecewise constraints are upper
        bounds then the piecewise constraints will be substituted for
        linear constraints. Setting 'force_pw=True' will force the use
        of the original piecewise constraints even when one of these two
        cases applies.

    warning_tol : float, default=1e-8
        To aid in debugging, a warning is printed when consecutive
        slopes of piecewise segments are within <warning_tol> of each
        other.

    warn_domain_coverage : bool, default=True
        Print a warning when the feasible region of the domain variable
        is not completely covered by the piecewise breakpoints.

    unbounded_domain_var : bool, default=False
        Allow an unbounded or partially bounded Pyomo Var to be used as
        the domain variable.

        .. note::
            This does not imply unbounded piecewise segments will be
            constructed. The outermost piecewise breakpoints will bound
            the domain variable at each index. However, the Var
            attributes .lb and .ub will not be modified.

    """

    _ComponentDataClass = PiecewiseData

    def __new__(cls, *args, **kwds):
        if cls != Piecewise:
            return super(Piecewise, cls).__new__(cls)
        if len(args) == 2:
            return SimplePiecewise.__new__(SimplePiecewise)
        else:
            return IndexedPiecewise.__new__(IndexedPiecewise)

    def __init__(self, *args, **kwds):
        # this is temporary as part of a move to user inputs
        # using Enums rather than strings
        translate_repn = {
            'BIGM_SOS1': PWRepn.BIGM_SOS1,
            PWRepn.BIGM_SOS1: PWRepn.BIGM_SOS1,
            'BIGM_BIN': PWRepn.BIGM_BIN,
            PWRepn.BIGM_BIN: PWRepn.BIGM_BIN,
            'SOS2': PWRepn.SOS2,
            PWRepn.SOS2: PWRepn.SOS2,
            'CC': PWRepn.CC,
            PWRepn.CC: PWRepn.CC,
            'DCC': PWRepn.DCC,
            PWRepn.DCC: PWRepn.DCC,
            'DLOG': PWRepn.DLOG,
            PWRepn.DLOG: PWRepn.DLOG,
            'LOG': PWRepn.LOG,
            PWRepn.LOG: PWRepn.LOG,
            'MC': PWRepn.MC,
            PWRepn.MC: PWRepn.MC,
            'INC': PWRepn.INC,
            PWRepn.INC: PWRepn.INC,
            None: None,
        }

        # this is temporary as part of a move to user inputs
        # using Enums rather than strings
        translate_bound = {
            'UB': Bound.Upper,
            Bound.Upper: Bound.Upper,
            'LB': Bound.Lower,
            Bound.Lower: Bound.Lower,
            'EQ': Bound.Equal,
            Bound.Equal: Bound.Equal,
            None: None,
        }

        # TODO: Update the keyword names. I think these are more clear
        #       pw_pts                  -> breakpoints
        #       pw_repn                 -> repn
        #       pw_constr_type          -> bound
        #       f_rule                  -> rule
        #       force_pw                -> simplify
        #       warning_tol
        #       warn_domain_coverage    -> warning_domain_coverage
        #       unbounded_domain_var
        #

        # extract all keywords used by this class
        pw_points = kwds.pop('pw_pts', None)
        # translate the user input to the enum type
        pw_rep = kwds.pop('pw_repn', 'SOS2')
        pw_rep = translate_repn.get(pw_rep, pw_rep)
        if (pw_rep == PWRepn.BIGM_BIN) or (pw_rep == PWRepn.BIGM_SOS1):
            deprecation_warning(
                "The 'BIGM_BIN' and 'BIGM_SOS1' "
                "piecewise representations will be removed in "
                "a future version of Pyomo. They produce incorrect "
                "results in certain cases",
                version='5.3',
            )
        # translate the user input to the enum type
        bound_type = kwds.pop('pw_constr_type', None)
        bound_type = translate_bound.get(bound_type, bound_type)
        f_rule = kwds.pop('f_rule', None)
        force_pw = kwds.pop('force_pw', False)
        warning_tol = kwds.pop('warning_tol', _WARNING_TOLERANCE)
        warn_domain_coverage = kwds.pop('warn_domain_coverage', True)
        unbounded_domain_var = kwds.pop('unbounded_domain_var', False)

        # all but the last two args should go to Block
        try:
            # Blocks have special handling when calling
            # __setattr__ with anything derived from component.
            # However, in this particular case we need to override
            # this so that these two variables don't get re-added
            # as new model components, therefore we directly modify
            # __dict__
            self.__dict__['_domain_var'] = args[-1]
            self.__dict__['_range_var'] = args[-2]
        except IndexError:
            msg = "Piecewise component initialized with less than two arguments"
            raise TypeError(msg)

        args = args[:-2]
        #
        # NOTE: The 'ctype' keyword argument is not defined here. This
        #       mocks what is done in PyomoModel.py, although here it
        #       feels like somewhat of a hack. The alternative is to
        #       modify the block_data_objects() method to include Piecewise
        #       type components. I'm not sure which is best at this
        #       time. Although, a consequence of the current
        #       implementation is that model.pprint() labels Piecewise
        #       blocks as simply Blocks.
        #
        # kwds.setdefault('ctype', Piecewise)
        Block.__init__(self, *args, **kwds)

        # Check that the variables args are actually Pyomo Vars
        if not (
            isinstance(self._domain_var, VarData)
            or isinstance(self._domain_var, IndexedVar)
        ):
            msg = (
                "Piecewise component has invalid "
                "argument type for domain variable, %s"
            )
            raise TypeError(msg % (repr(self._domain_var),))
        if not (
            isinstance(self._range_var, VarData)
            or isinstance(self._range_var, IndexedVar)
        ):
            msg = (
                "Piecewise component has invalid "
                "argument type for range variable, %s"
            )
            raise TypeError(msg % (repr(self._range_var),))

        # Test that the keyword values make sense
        if f_rule.__class__ not in [type(lambda: None), dict, list, tuple]:
            msg = (
                "Piecewise component keyword 'f_rule' must "
                "be a function, dict, list, or tuple"
            )
            raise ValueError(msg)
        try:
            bound_type = Bound(bound_type)
        except ValueError:
            try:
                bound_type = Bound[bound_type]
            except KeyError:
                msg = "Invalid value for Piecewise component keyword 'pw_constr_type'"
                raise ValueError(msg)
        if warning_tol.__class__ is not float:
            msg = (
                "Invalid type '%s' for Piecewise component "
                "keyword 'warning_tol', which must be of type 'float'"
            )
            raise TypeError(msg % (type(warning_tol),))
        if warn_domain_coverage not in [True, False]:
            msg = (
                "Invalid value for Piecewise component "
                "keyword 'warn_domain_coverage', which must be True or False"
            )
            raise ValueError(msg)
        if unbounded_domain_var not in [True, False]:
            msg = (
                "Invalid value for Piecewise component "
                "keyword 'unbounded_domain_var', which must be True or False"
            )
            raise ValueError(msg)

        self._pw_rep = pw_rep
        self._bound_type = bound_type
        self._f_rule = f_rule
        self._force_pw = force_pw
        self._warning_tol = warning_tol
        self._warn_domain_coverage = warn_domain_coverage
        self._unbounded_domain_var = unbounded_domain_var

        if self.is_indexed() is False:
            if not (isinstance(pw_points, list) or isinstance(pw_points, tuple)):
                msg = (
                    "Invalid type '%s' for Piecewise component "
                    "keyword 'pw_pts', which must be of type "
                    "'list' or 'tuple' for non-indexed Piecewise component"
                )
                raise TypeError(msg % (type(pw_points),))
            self._domain_points = {None: pw_points}
        else:
            if isinstance(pw_points, list) or isinstance(pw_points, tuple):
                self._domain_points = {None: pw_points}
            elif isinstance(pw_points, dict):
                self._domain_points = pw_points
            else:
                msg = (
                    "Invalid type '%s' for Piecewise component "
                    "keyword 'pw_pts', which must be of type "
                    "'dict', 'list', or 'tuple' for indexed Piecewise component"
                )
                raise TypeError(msg % (type(pw_points),))

    def construct(self, *args, **kwds):
        """
        A quick hack to call add after data has been loaded.
        """
        generate_debug_messages = is_debug_set(logger)

        if self._constructed:
            return
        timer = ConstructionTimer(self)
        # We need to be able to add and construct new model
        # components on the fly so we make this Block behave concretely
        self._constructed = True

        # cache this because it is apparently expensive
        is_indexed = self.is_indexed()

        # construct each index of this component
        if not is_indexed:
            if generate_debug_messages:
                logger.debug("  Constructing single Piecewise component (index=None)")
            self.add(None, _is_indexed=is_indexed)
        else:
            for index in self._index_set:
                if generate_debug_messages:
                    logger.debug("  Constructing Piecewise index " + str(index))
                self.add(index, _is_indexed=is_indexed)
        timer.report()

    def add(self, index, _is_indexed=None):
        if _is_indexed is None:
            _is_indexed = self.is_indexed()

        _self_parent = self._parent()
        _self_xvar = None
        _self_yvar = None
        _self_domain_pts_index = None
        if not _is_indexed:
            # allows one to mix Var and VarData as input to
            # non-indexed Piecewise, index would be None in this case
            # so for Var elements Var[None] is Var, but VarData[None] would fail
            _self_xvar = self._domain_var
            _self_yvar = self._range_var
            _self_domain_pts_index = self._domain_points[index]
        else:
            # The following allows one to specify a Var or VarData
            # object even with an indexed Piecewise component.
            # The most common situation will most likely be a VarArray,
            # so we try this first.
            if not isinstance(self._domain_var, VarData):
                _self_xvar = self._domain_var[index]
            else:
                _self_xvar = self._domain_var
            if not isinstance(self._range_var, VarData):
                _self_yvar = self._range_var[index]
            else:
                _self_yvar = self._range_var
            try:
                _self_domain_pts_index = self._domain_points[index]
            except KeyError:
                # This hack was set up in __init__ for using a single list or tuple
                # with an indexed piecewise constraint. I assume this will fail if
                # None is ever allowed as a valid index for an IndexedComponent,
                # hence the assert below
                assert not (_is_indexed and (index is None))
                _self_domain_pts_index = self._domain_points[None]

        if self._unbounded_domain_var is False:
            # We add the requirement that the domain variable used by Piecewise is
            # always bounded from above and below.
            if (_self_xvar.lb is None) or (_self_xvar.ub is None):
                msg = (
                    "Piecewise '%s[%s]' found an unbounded variable "
                    "used for the constraint domain: '%s'. "
                    "Piecewise component requires the domain variable have "
                    "lower and upper bounds. Refer to the Piecewise help "
                    "documentation for information on how to disable this "
                    "restriction"
                )
                raise ValueError(msg % (self.name, index, _self_xvar))

        if self._warn_domain_coverage is True:
            # Print a warning when the feasible region created by the piecewise
            # constraints does not include the domain variables bounds
            if (_self_xvar.lb is not None) and (
                _self_xvar.lb < min(_self_domain_pts_index)
            ):
                msg = (
                    "**WARNING: Piecewise '%s[%s]' feasible region does not "
                    "include the lower bound of domain variable: %s.lb = %s < %s. "
                    "Refer to the Piecewise help documentation for information on "
                    "how to disable this warning."
                )
                print(
                    msg
                    % (
                        self.name,
                        index,
                        _self_xvar,
                        _self_xvar.lb,
                        min(_self_domain_pts_index),
                    )
                )
            if (_self_xvar.ub is not None) and (
                _self_xvar.ub > max(_self_domain_pts_index)
            ):
                msg = (
                    "**WARNING: Piecewise '%s[%s]' feasible region does not "
                    "include the upper bound of domain variable: %s.ub = %s > %s. "
                    "Refer to the Piecewise help documentation for information on "
                    "how to disable this warning."
                )
                print(
                    msg
                    % (
                        self.name,
                        index,
                        _self_xvar,
                        _self_xvar.ub,
                        max(_self_domain_pts_index),
                    )
                )

        if len(_self_domain_pts_index) <= 1:
            # TODO: Technically one could interpret this
            #       case by adding simple constraints that
            #       fix the domain and range variable to the
            #       single (x,y) point that is given. This
            #       seems like it would be a bug more often
            #       than not, so I don't believe it should
            #       be the default behavior.
            raise ValueError(
                "Piecewise component '%s[%s]' failed to construct "
                "piecewise representation. List of breakpoints "
                "must contain at least two elements. Current list: %s"
                % (self.name, index, str(_self_domain_pts_index))
            )

        # generate the list of range values using the function rule
        # check if convexity or concavity holds as well
        force_simple = False
        if not _is_indexed:
            character, range_pts, isStep = _characterize_function(
                self.name,
                self._warning_tol,
                self._f_rule,
                _self_parent,
                _self_domain_pts_index,
            )
        else:
            character, range_pts, isStep = _characterize_function(
                self.name,
                self._warning_tol,
                self._f_rule,
                _self_parent,
                _self_domain_pts_index,
                index,
            )

        assert not ((isStep) and (character in [-1, 1]))
        if (isStep) and (
            self._pw_rep in [PWRepn.MC, PWRepn.BIGM_BIN, PWRepn.BIGM_SOS1]
        ):
            msg = (
                "Piecewise '%s[%s]' has detected a step function but the selected "
                "piecewise representation '%s' does not currently support this "
                "functionality. Refer to the Piecewise help documentation for "
                "information about which piecewise representations support step functions."
            )
            raise ValueError(msg % (self.name, index, self._pw_rep))

        # Make automatic simplifications to the piecewise constraints
        # for the special cases of convexity and lower bound
        # or concavity and upper bound
        if character == -1:
            if self._bound_type == Bound.Upper:
                force_simple = True
        elif character == 1:
            if self._bound_type == Bound.Lower:
                force_simple = True

        # make sure the user does not want to disable the automatic
        # simplifications above
        if self._force_pw is True:
            force_simple = False

        func = None
        if force_simple is True:
            # In the case where the feasible region is convex or
            # concave (and the user does not want to force the use of
            # piecewise constraints) use simple linear constraints.
            func = _SimplifiedPiecewise()
        else:
            if len(_self_domain_pts_index) == 2:
                # Always use a simple single line constraint when
                # only two points are present in the piecewise list
                func = _SimpleSinglePiecewise()
            else:
                # generate piecewise constraints
                if self._pw_rep == PWRepn.SOS2:
                    func = _SOS2Piecewise()
                elif self._pw_rep == PWRepn.INC:
                    func = _INCPiecewise()
                elif self._pw_rep == PWRepn.MC:
                    func = _MCPiecewise()
                elif self._pw_rep == PWRepn.DCC:
                    func = _DCCPiecewise()
                elif self._pw_rep == PWRepn.DLOG:
                    func = _DLOGPiecewise()
                elif self._pw_rep == PWRepn.CC:
                    func = _CCPiecewise()
                elif self._pw_rep == PWRepn.LOG:
                    func = _LOGPiecewise()
                elif self._pw_rep == PWRepn.BIGM_BIN:
                    func = _BIGMPiecewise(binary=True)
                elif self._pw_rep == PWRepn.BIGM_SOS1:
                    func = _BIGMPiecewise(binary=False)
                else:
                    msg = (
                        "Piecewise '%s[%s]' does not have a valid "
                        "piecewise representation: '%s'"
                    )
                    raise ValueError(msg % (self.name, index, self._pw_rep))

        if _is_indexed:
            comp = PiecewiseData(self)
        else:
            comp = self
        self._data[index] = comp
        comp._index = index
        comp.updateBoundType(self._bound_type)
        comp.updatePoints(_self_domain_pts_index, range_pts)
        comp.build_constraints(func, _self_xvar, _self_yvar)


class SimplePiecewise(PiecewiseData, Piecewise):
    def __init__(self, *args, **kwds):
        PiecewiseData.__init__(self, self)
        Piecewise.__init__(self, *args, **kwds)


class IndexedPiecewise(Piecewise):
    def __init__(self, *args, **kwds):
        Piecewise.__init__(self, *args, **kwds)

    def __str__(self):
        return str(self.name)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""
Rooney Biegler DoE example, based on Rooney, W. C. and Biegler, L. T. (2001). Design for
model parameter uncertainty using nonlinear confidence regions. AIChE Journal,
47(8), 1794-1804.
"""

from pyomo.contrib.parmest.examples.rooney_biegler.rooney_biegler import (
    RooneyBieglerExperiment,
)
from pyomo.contrib.doe import DesignOfExperiments, ObjectiveLib
from pyomo.common.dependencies import pandas as pd, numpy as np, matplotlib


def run_rooney_biegler_doe(
    sensitivity_formula="central",
    optimization_objective="determinant",
    improve_cholesky_roundoff_error=False,
    compute_FIM_full_factorial=False,
    draw_factorial_figure=False,
    design_range=None,
    tee=False,
    print_output=False,
):
    """Run Rooney Biegler DoE Example. All the boolean options control whether
    that section of the example is run. This is to facilitate testing using the example.

    Parameters
    ----------
    sensitivity_formula : str, optional
        Differentiation formula for computing the sensitivity matrix.
        Must be one of ['central', 'forward', 'backward']. By default 'central'.
        Note: symbolic differentiation is not currently supported for this example.
        Can be added in the future when that is implemented in Pyomo.DoE.
    optimization_objective : str or ObjectiveLib, optional
        Objective function for the design of experiments optimization.
        Can be a string ('determinant', 'trace', 'pseudo_trace', 'minimum_eigenvalue',
        'condition_number') or an ObjectiveLib enum member. By default 'determinant'.
    improve_cholesky_roundoff_error : bool, optional
        If True, applies additional constraints to improve Cholesky factorization
        round-off error. By default False.
    compute_FIM_full_factorial : bool, optional
        If True, computes the Fisher Information Matrix for all combinations in the
        full factorial design space defined by design_range. By default False.
    draw_factorial_figure : bool, optional
        If True, generates and saves factorial figures using the data from
        `compute_FIM_full_factorial` showing the design space exploration.
        By default False.
    design_range : dict
        Dictionary specifying the range of design variables to explore. Keys are
        variable names and values are lists of candidate values. Required for
        `compute_FIM_full_factorial`. By default {'hour': [0, 10, 40]}.
    tee : bool, optional
        If True, displays solver output during optimization. By default False.
    print_output : bool, optional
        If True, prints optimization results and full factorial design results
        to the console. By default False.

    Returns
    -------
    dict
        A dictionary containing:
        - 'experiment': The RooneyBieglerExperiment object used for design
        - 'results_dict': Full factorial design results (if computed)
        - 'optimization': Dictionary with optimization results including objective
          values and optimal designs
    """
    # Convert optimization_objective to enum if it's a string
    if isinstance(optimization_objective, str):
        optimization_objective = ObjectiveLib(optimization_objective)

    if design_range is None:
        design_range = {'hour': [0, 10, 40]}

    # Initialize a container for all potential results
    results_container = {
        "experiment": None,
        "results_dict": {},
        "optimization": {},  # Will hold D/A optimization results if run
    }

    # Data Setup
    data = pd.DataFrame(
        data=[[1, 8.3], [2, 10.3], [3, 19.0], [4, 16.0], [5, 15.6], [7, 19.8]],
        columns=['hour', 'y'],
    )
    theta = {'asymptote': 15, 'rate_constant': 0.5}
    measurement_error = 0.1

    # Compute prior FIM from existing data
    FIM = np.zeros((2, 2))
    for i in range(len(data)):
        exp_i = RooneyBieglerExperiment(
            data=data.loc[i, :], theta=theta, measure_error=measurement_error
        )
        doe_prior = DesignOfExperiments(
            experiment=exp_i,
            fd_formula=sensitivity_formula,
            objective_option=optimization_objective,
            tee=tee,
        )
        FIM += doe_prior.compute_FIM()

    # Base Experiment for Design
    rooney_biegler_experiment = RooneyBieglerExperiment(
        data=data.loc[0, :], theta=theta, measure_error=measurement_error
    )
    results_container["experiment"] = rooney_biegler_experiment

    rooney_biegler_doe = DesignOfExperiments(
        experiment=rooney_biegler_experiment,
        objective_option=optimization_objective,
        fd_formula=sensitivity_formula,
        tee=tee,
        prior_FIM=FIM,
        improve_cholesky_roundoff_error=improve_cholesky_roundoff_error,
    )

    # Run the optimization
    rooney_biegler_doe.run_doe()

    # Store results based on objective type
    objective_name = optimization_objective.value
    results_container["optimization"] = {
        "objective_type": objective_name,
        "design": rooney_biegler_doe.results['Experiment Design'],
    }

    # Add objective-specific metrics
    if optimization_objective == ObjectiveLib.determinant:
        results_container["optimization"]["value"] = rooney_biegler_doe.results[
            'log10 D-opt'
        ]
    elif optimization_objective == ObjectiveLib.trace:
        results_container["optimization"]["value"] = rooney_biegler_doe.results[
            'log10 A-opt'
        ]
    elif optimization_objective == ObjectiveLib.pseudo_trace:
        results_container["optimization"]["value"] = rooney_biegler_doe.results[
            'log10 pseudo A-opt'
        ]
    elif optimization_objective == ObjectiveLib.minimum_eigenvalue:
        results_container["optimization"]["value"] = rooney_biegler_doe.results[
            'log10 E-opt'
        ]
    elif optimization_objective == ObjectiveLib.condition_number:
        results_container["optimization"]["value"] = rooney_biegler_doe.results[
            'FIM Condition Number'
        ]

    if print_output:
        print(f"Optimal results for {objective_name}:", rooney_biegler_doe.results)

    # Compute Full Factorial Design Results
    if compute_FIM_full_factorial:
        results_container["results_dict"] = (
            rooney_biegler_doe.compute_FIM_full_factorial(design_ranges=design_range)
        )
        if print_output:
            print("Full Factorial Design Results:\n", results_container["results_dict"])

    # Pyomo.DoE built-in factorial figure generation
    if draw_factorial_figure:
        rooney_biegler_doe.draw_factorial_figure(
            sensitivity_design_variables=['hour'],
            fixed_design_variables={},
            title_text="Rooney Biegler DoE - Full Factorial Design",
            xlabel_text="Experiment duration (hour)",
            log_scale=False,
            figure_file_name="rooney_biegler",
        )

    return results_container


if __name__ == "__main__":

    results = run_rooney_biegler_doe(
        optimization_objective=ObjectiveLib.trace,  # Can also use string: "trace"
        sensitivity_formula="central",
        improve_cholesky_roundoff_error=False,
        compute_FIM_full_factorial=True,
        draw_factorial_figure=True,  # Set True to test file generation
        design_range={'hour': [0, 10, 3]},  # Small range for speed
        print_output=True,
    )
    print(results)
    # Show plots if running locally
    matplotlib.pyplot.show()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

# -*- coding: utf-8 -*-
"""Tests for the MindtPy solver."""

import pyomo.common.unittest as unittest
from pyomo.contrib.mindtpy.tests.eight_process_problem import EightProcessFlowsheet
from pyomo.contrib.mindtpy.tests.constraint_qualification_example import (
    ConstraintQualificationExample,
)
from pyomo.environ import SolverFactory, value
from pyomo.opt import TerminationCondition

required_solvers = ('ipopt', 'cplex')
# required_solvers = ('gams', 'gams')
if all(SolverFactory(s).available(exception_flag=False) for s in required_solvers):
    subsolvers_available = True
else:
    subsolvers_available = False


model_list = [EightProcessFlowsheet(convex=True), ConstraintQualificationExample()]


@unittest.skipIf(
    not subsolvers_available,
    'Required subsolvers %s are not available' % (required_solvers,),
)
class TestMindtPy(unittest.TestCase):
    """Tests for the MindtPy solver plugin."""

    def check_optimal_solution(self, model, places=1):
        for var in model.optimal_solution:
            self.assertAlmostEqual(
                var.value, model.optimal_solution[var], places=places
            )

    def test_ROA_L1(self):
        """Test the LP/NLP decomposition algorithm."""
        with SolverFactory('mindtpy') as opt:
            for model in model_list:
                model = model.clone()
                results = opt.solve(
                    model,
                    strategy='OA',
                    mip_solver=required_solvers[1],
                    nlp_solver=required_solvers[0],
                    add_regularization='level_L1',
                )

                self.assertIn(
                    results.solver.termination_condition,
                    [TerminationCondition.optimal, TerminationCondition.feasible],
                )
                self.assertAlmostEqual(
                    value(model.objective.expr), model.optimal_value, places=1
                )
                self.check_optimal_solution(model)

    def test_ROA_L2(self):
        """Test the LP/NLP decomposition algorithm."""
        with SolverFactory('mindtpy') as opt:
            for model in model_list:
                model = model.clone()
                results = opt.solve(
                    model,
                    strategy='OA',
                    mip_solver=required_solvers[1],
                    nlp_solver=required_solvers[0],
                    add_regularization='level_L2',
                )

                self.assertIn(
                    results.solver.termination_condition,
                    [TerminationCondition.optimal, TerminationCondition.feasible],
                )
                self.assertAlmostEqual(
                    value(model.objective.expr), model.optimal_value, places=1
                )
                self.check_optimal_solution(model)

    def test_ROA_Linf(self):
        """Test the LP/NLP decomposition algorithm."""
        with SolverFactory('mindtpy') as opt:
            for model in model_list:
                model = model.clone()
                results = opt.solve(
                    model,
                    strategy='OA',
                    mip_solver=required_solvers[1],
                    nlp_solver=required_solvers[0],
                    add_regularization='level_L_infinity',
                )

                self.assertIn(
                    results.solver.termination_condition,
                    [TerminationCondition.optimal, TerminationCondition.feasible],
                )
                self.assertAlmostEqual(
                    value(model.objective.expr), model.optimal_value, places=1
                )
                self.check_optimal_solution(model)

    def test_ROA_grad_lag(self):
        """Test the LP/NLP decomposition algorithm."""
        with SolverFactory('mindtpy') as opt:
            for model in model_list:
                model = model.clone()
                results = opt.solve(
                    model,
                    strategy='OA',
                    mip_solver=required_solvers[1],
                    nlp_solver=required_solvers[0],
                    add_regularization='grad_lag',
                )

                self.assertIn(
                    results.solver.termination_condition,
                    [TerminationCondition.optimal, TerminationCondition.feasible],
                )
                self.assertAlmostEqual(
                    value(model.objective.expr), model.optimal_value, places=1
                )
                self.check_optimal_solution(model)

    def test_ROA_hess_lag(self):
        """Test the LP/NLP decomposition algorithm."""
        with SolverFactory('mindtpy') as opt:
            for model in model_list:
                model = model.clone()
                results = opt.solve(
                    model,
                    strategy='OA',
                    mip_solver=required_solvers[1],
                    nlp_solver=required_solvers[0],
                    add_regularization='hess_lag',
                )

                self.assertIn(
                    results.solver.termination_condition,
                    [TerminationCondition.optimal, TerminationCondition.feasible],
                )
                self.assertAlmostEqual(
                    value(model.objective.expr), model.optimal_value, places=1
                )
                self.check_optimal_solution(model)

    def test_ROA_hess_only_lag(self):
        """Test the LP/NLP decomposition algorithm."""
        with SolverFactory('mindtpy') as opt:
            for model in model_list:
                model = model.clone()
                results = opt.solve(
                    model,
                    strategy='OA',
                    mip_solver=required_solvers[1],
                    nlp_solver=required_solvers[0],
                    add_regularization='hess_only_lag',
                )

                self.assertIn(
                    results.solver.termination_condition,
                    [TerminationCondition.optimal, TerminationCondition.feasible],
                )
                self.assertAlmostEqual(
                    value(model.objective.expr), model.optimal_value, places=1
                )
                self.check_optimal_solution(model)

    def test_ROA_sqp_lag(self):
        """Test the LP/NLP decomposition algorithm."""
        with SolverFactory('mindtpy') as opt:
            for model in model_list:
                model = model.clone()
                results = opt.solve(
                    model,
                    strategy='OA',
                    mip_solver=required_solvers[1],
                    nlp_solver=required_solvers[0],
                    add_regularization='sqp_lag',
                )

                self.assertIn(
                    results.solver.termination_condition,
                    [TerminationCondition.optimal, TerminationCondition.feasible],
                )
                self.assertAlmostEqual(
                    value(model.objective.expr), model.optimal_value, places=1
                )
                self.check_optimal_solution(model)

    def test_ROA_sqp_lag_equality_relaxation(self):
        """Test the LP/NLP decomposition algorithm."""
        with SolverFactory('mindtpy') as opt:
            for model in model_list:
                model = model.clone()
                results = opt.solve(
                    model,
                    strategy='OA',
                    mip_solver=required_solvers[1],
                    nlp_solver=required_solvers[0],
                    add_regularization='sqp_lag',
                    equality_relaxation=True,
                )

                self.assertIn(
                    results.solver.termination_condition,
                    [TerminationCondition.optimal, TerminationCondition.feasible],
                )
                self.assertAlmostEqual(
                    value(model.objective.expr), model.optimal_value, places=1
                )
                self.check_optimal_solution(model)

    def test_ROA_sqp_lag_add_no_good_cuts(self):
        """Test the LP/NLP decomposition algorithm."""
        with SolverFactory('mindtpy') as opt:
            for model in model_list:
                model = model.clone()
                results = opt.solve(
                    model,
                    strategy='OA',
                    mip_solver=required_solvers[1],
                    nlp_solver=required_solvers[0],
                    add_regularization='sqp_lag',
                    equality_relaxation=True,
                    add_no_good_cuts=True,
                )

                self.assertIn(
                    results.solver.termination_condition,
                    [TerminationCondition.optimal, TerminationCondition.feasible],
                )
                self.assertAlmostEqual(
                    value(model.objective.expr), model.optimal_value, places=1
                )
                self.check_optimal_solution(model)

    def test_ROA_sqp_lag_level_coef(self):
        """Test the LP/NLP decomposition algorithm."""
        with SolverFactory('mindtpy') as opt:
            for model in model_list:
                model = model.clone()
                results = opt.solve(
                    model,
                    strategy='OA',
                    mip_solver=required_solvers[1],
                    nlp_solver=required_solvers[0],
                    add_regularization='sqp_lag',
                    equality_relaxation=True,
                    level_coef=0.4,
                )

                self.assertIn(
                    results.solver.termination_condition,
                    [TerminationCondition.optimal, TerminationCondition.feasible],
                )
                self.assertAlmostEqual(
                    value(model.objective.expr), model.optimal_value, places=1
                )
                self.check_optimal_solution(model)


if __name__ == '__main__':
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from collections import namedtuple
from heapq import heappush, heappop
import traceback

from pyomo.common.collections import ComponentMap
from pyomo.common.config import document_kwargs_from_configdict
from pyomo.common.errors import InfeasibleConstraintException
from pyomo.contrib.fbbt.fbbt import fbbt
from pyomo.contrib.gdpopt.algorithm_base_class import _GDPoptAlgorithm
from pyomo.contrib.gdpopt.create_oa_subproblems import (
    add_util_block,
    add_disjunction_list,
    add_disjunct_list,
    add_algebraic_variable_list,
    add_boolean_variable_lists,
    add_transformed_boolean_variable_list,
)
from pyomo.contrib.gdpopt.config_options import (
    _add_nlp_solver_configs,
    _add_BB_configs,
    _add_mip_solver_configs,
    _add_tolerance_configs,
    _add_nlp_solve_configs,
)
from pyomo.contrib.gdpopt.nlp_initialization import restore_vars_to_original_values
from pyomo.contrib.gdpopt.util import (
    copy_var_list_values,
    SuppressInfeasibleWarning,
    get_main_elapsed_time,
)
from pyomo.contrib.satsolver.satsolver import satisfiable
from pyomo.core import minimize, Suffix, Constraint, TransformationFactory
from pyomo.opt import SolverFactory, SolverStatus
from pyomo.opt import TerminationCondition as tc

_linear_degrees = {1, 0}

# Data tuple for each node that also functions as the sort key.
# Therefore, ordering of the arguments below matters.
BBNodeData = namedtuple(
    'BBNodeData',
    [
        'obj_lb',  # lower bound on objective value, sign corrected to minimize
        'obj_ub',  # upper bound on objective value, sign corrected to minimize
        'is_screened',  # True if the node has been screened; False if not.
        'is_evaluated',  # True if node has been evaluated; False if not.
        'num_unbranched_disjunctions',  # number of unbranched disjunctions
        'node_count',  # cumulative node counter
        'unbranched_disjunction_indices',  # list of unbranched disjunction indices
    ],
)


@SolverFactory.register(
    'gdpopt.lbb',
    doc="The LBB (logic-based branch and bound) Generalized Disjunctive "
    "Programming (GDP) solver",
)
class GDP_LBB_Solver(_GDPoptAlgorithm):
    """The GDPopt (Generalized Disjunctive Programming optimizer) logic-based
    branch and bound (LBB) solver.

    Accepts models that can include nonlinear, continuous variables and
    constraints, as well as logical conditions.
    """

    CONFIG = _GDPoptAlgorithm.CONFIG()
    _add_mip_solver_configs(CONFIG)
    _add_nlp_solver_configs(CONFIG, default_solver='ipopt')
    _add_nlp_solve_configs(
        CONFIG, default_nlp_init_method=restore_vars_to_original_values
    )
    _add_tolerance_configs(CONFIG)
    _add_BB_configs(CONFIG)

    algorithm = 'LBB'

    # Override solve() to customize the docstring for this solver
    @document_kwargs_from_configdict(CONFIG, doc=_GDPoptAlgorithm.solve.__doc__)
    def solve(self, model, **kwds):
        return super().solve(model, **kwds)

    def _log_citation(self, config):
        config.logger.info("\n" + """- LBB algorithm:
        Lee, S; Grossmann, IE.
        New algorithms for nonlinear generalized disjunctive programming.
        Comp. and Chem. Eng. 2000, 24, 2125-2141.
        DOI: 10.1016/S0098-1354(00)00581-0.
        """.strip())

    def _solve_gdp(self, model, config):
        self.explored_nodes = 0

        # Create utility block on the original model so that we will be able to
        # copy solutions between
        util_block = self.original_util_block = add_util_block(model)
        add_disjunct_list(util_block)
        add_algebraic_variable_list(util_block)
        add_boolean_variable_lists(util_block)

        root_node = TransformationFactory(
            'contrib.logical_to_disjunctive'
        ).create_using(model)
        root_util_blk = root_node.component(self.original_util_block.name)
        # Add to root utility block what we will need during the algorithm
        add_disjunction_list(root_util_blk)
        # Now that logical_to_disjunctive has been called.
        add_transformed_boolean_variable_list(root_util_blk)

        # Map unfixed disjunct -> list of deactivated constraints
        root_util_blk.disjunct_to_nonlinear_constraints = ComponentMap()
        # Map relaxed disjunctions -> list of unfixed disjuncts
        root_util_blk.disjunction_to_unfixed_disjuncts = ComponentMap()

        # Preprocess the active disjunctions
        for disjunction in root_util_blk.disjunction_list:
            assert disjunction.active

            disjuncts_fixed_True = []
            disjuncts_fixed_False = []
            unfixed_disjuncts = []

            # categorize the disjuncts in the disjunction
            for disjunct in disjunction.disjuncts:
                if disjunct.indicator_var.fixed:
                    if disjunct.indicator_var.value:
                        disjuncts_fixed_True.append(disjunct)
                    elif disjunct.indicator_var.value == False:
                        disjuncts_fixed_False.append(disjunct)
                else:
                    unfixed_disjuncts.append(disjunct)

            # update disjunct lists for predetermined disjunctions
            if len(disjuncts_fixed_False) == len(disjunction.disjuncts) - 1:
                # all but one disjunct in the disjunction is fixed to False.
                # Remaining one must be true. If not already fixed to True, do
                # so.
                if not disjuncts_fixed_True:
                    disjuncts_fixed_True = unfixed_disjuncts[0]
                    unfixed_disjuncts = []
                    disjuncts_fixed_True.indicator_var.fix(True)
            elif disjuncts_fixed_True and disjunction.xor:
                assert len(disjuncts_fixed_True) == 1, (
                    "XOR (only one True) violated: %s" % disjunction.name
                )
                disjuncts_fixed_False.extend(unfixed_disjuncts)
                unfixed_disjuncts = []

            # Make sure disjuncts fixed to False are properly deactivated.
            for disjunct in disjuncts_fixed_False:
                disjunct.deactivate()

            # Deactivate nonlinear constraints in unfixed disjuncts
            for disjunct in unfixed_disjuncts:
                nonlinear_constraints_in_disjunct = [
                    constr
                    for constr in disjunct.component_data_objects(
                        Constraint, active=True
                    )
                    if constr.body.polynomial_degree() not in _linear_degrees
                ]
                for constraint in nonlinear_constraints_in_disjunct:
                    constraint.deactivate()
                if nonlinear_constraints_in_disjunct:
                    # TODO might be worthwhile to log number of nonlinear
                    # constraints in each disjunction for later branching
                    # purposes
                    root_util_blk.disjunct_to_nonlinear_constraints[disjunct] = (
                        nonlinear_constraints_in_disjunct
                    )

            root_util_blk.disjunction_to_unfixed_disjuncts[disjunction] = (
                unfixed_disjuncts
            )
            pass

        # Add the BigM suffix if it does not already exist. Used later during
        # nonlinear constraint activation.
        if not hasattr(root_node, 'BigM'):
            root_node.BigM = Suffix()

        # Set up the priority queue
        queue = self.bb_queue = []
        self.created_nodes = 0
        unbranched_disjunction_indices = [
            i
            for i, disjunction in enumerate(root_util_blk.disjunction_list)
            if disjunction in root_util_blk.disjunction_to_unfixed_disjuncts
        ]
        sort_tuple = BBNodeData(
            obj_lb=float('-inf'),
            obj_ub=float('inf'),
            is_screened=False,
            is_evaluated=False,
            num_unbranched_disjunctions=len(unbranched_disjunction_indices),
            node_count=0,
            unbranched_disjunction_indices=unbranched_disjunction_indices,
        )
        heappush(queue, (sort_tuple, root_node))

        # Do the branch and bound
        while len(queue) > 0:
            # visit the top node on the heap
            node_data, node_model = heappop(queue)
            config.logger.info(
                "Nodes: %s LB %.10g Unbranched %s"
                % (
                    self.explored_nodes,
                    node_data.obj_lb,
                    node_data.num_unbranched_disjunctions,
                )
            )

            # Check time limit
            if self.reached_time_limit(config):
                no_feasible_soln = float('inf')
                self.LB = (
                    node_data.obj_lb
                    if self.objective_sense == minimize
                    else -no_feasible_soln
                )
                self.UB = (
                    no_feasible_soln
                    if self.objective_sense == minimize
                    else -node_data.obj_lb
                )
                config.logger.info(
                    'Final bound values: LB: {}  UB: {}'.format(self.LB, self.UB)
                )
                return self._get_final_results_object()

            # Handle current node
            if not node_data.is_screened:
                # Node has not been evaluated.
                self.explored_nodes += 1
                new_node_data = self._prescreen_node(node_data, node_model, config)
                # replace with updated node data
                heappush(queue, (new_node_data, node_model))
            elif (
                node_data.obj_lb < node_data.obj_ub - config.bound_tolerance
                and not node_data.is_evaluated
            ):
                # Node has not been fully evaluated.
                # Note: infeasible and unbounded nodes will skip this condition,
                # because of strict inequality
                new_node_data = self._evaluate_node(node_data, node_model, config)
                # replace with updated node data
                heappush(queue, (new_node_data, node_model))
            elif (
                node_data.num_unbranched_disjunctions == 0
                or node_data.obj_lb == float('inf')
            ):
                # We have reached a leaf node, or the best available node is
                # infeasible.

                # Update the incumbent and put it in the original model
                self.update_incumbent(
                    node_model.component(self.original_util_block.name)
                )
                self._transfer_incumbent_to_original_model(config.logger)

                self.LB = (
                    node_data.obj_lb
                    if self.objective_sense == minimize
                    else -node_data.obj_ub
                )
                self.UB = (
                    node_data.obj_ub
                    if self.objective_sense == minimize
                    else -node_data.obj_lb
                )
                self.iteration = self.explored_nodes
                if node_data.obj_lb == float('inf'):
                    self.pyomo_results.solver.termination_condition = tc.infeasible
                elif node_data.obj_ub == float('-inf'):
                    self.pyomo_results.solver.termination_condition = tc.unbounded
                else:
                    self.pyomo_results.solver.termination_condition = tc.optimal
                return self._get_final_pyomo_results_object()
            else:
                self._branch_on_node(node_data, node_model, config)

    def _branch_on_node(self, node_data, node_model, config):
        node_utils = node_model.component(self.original_util_block.name)

        # Keeping the naive branch selection
        disjunction_to_branch_idx = node_data.unbranched_disjunction_indices[0]
        disjunction_to_branch = node_utils.disjunction_list[disjunction_to_branch_idx]
        num_unfixed_disjuncts = len(
            node_utils.disjunction_to_unfixed_disjuncts[disjunction_to_branch]
        )
        config.logger.info("Branching on disjunction %s" % disjunction_to_branch.name)
        node_count = self.created_nodes
        newly_created_nodes = 0

        for disjunct_index_to_fix_True in range(num_unfixed_disjuncts):
            # Create a new branch for each unfixed disjunct
            child_model = node_model.clone()
            child_utils = child_model.component(node_utils.name)
            child_disjunction_to_branch = child_utils.disjunction_list[
                disjunction_to_branch_idx
            ]
            child_unfixed_disjuncts = child_utils.disjunction_to_unfixed_disjuncts[
                child_disjunction_to_branch
            ]
            for idx, child_disjunct in enumerate(child_unfixed_disjuncts):
                if idx == disjunct_index_to_fix_True:
                    child_disjunct.indicator_var.fix(True)
                else:
                    child_disjunct.deactivate()
            if not child_disjunction_to_branch.xor:
                raise NotImplementedError(
                    "We still need to add support for non-XOR disjunctions."
                )
            # This requires adding all combinations of activation status among
            # unfixed_disjuncts Reactivate nonlinear constraints in the
            # newly-fixed child disjunct
            fixed_True_disjunct = child_unfixed_disjuncts[disjunct_index_to_fix_True]
            for constr in child_utils.disjunct_to_nonlinear_constraints.get(
                fixed_True_disjunct, ()
            ):
                constr.activate()
                child_model.BigM[constr] = 1  # set arbitrary BigM (ok, because
                # we fix corresponding Y=True)

            del child_utils.disjunction_to_unfixed_disjuncts[
                child_disjunction_to_branch
            ]
            for child_disjunct in child_unfixed_disjuncts:
                child_utils.disjunct_to_nonlinear_constraints.pop(child_disjunct, None)

            newly_created_nodes += 1
            child_node_data = node_data._replace(
                is_screened=False,
                is_evaluated=False,
                num_unbranched_disjunctions=node_data.num_unbranched_disjunctions - 1,
                node_count=node_count + newly_created_nodes,
                unbranched_disjunction_indices=node_data.unbranched_disjunction_indices[
                    1:
                ],
                obj_ub=float('inf'),
            )
            heappush(self.bb_queue, (child_node_data, child_model))

        self.created_nodes += newly_created_nodes

        config.logger.info(
            "Added %s new nodes with %s relaxed disjunctions to "
            "the heap. Size now %s."
            % (
                num_unfixed_disjuncts,
                node_data.num_unbranched_disjunctions - 1,
                len(self.bb_queue),
            )
        )

    def _prescreen_node(self, node_data, node_model, config):
        # Check node for satisfiability if sat-solver is enabled
        if config.check_sat and satisfiable(node_model, config.logger) is False:
            if node_data.node_count == 0:
                config.logger.info(
                    "Root node is not satisfiable. Problem is infeasible."
                )
            else:
                config.logger.info("SAT solver pruned node %s" % node_data.node_count)
            new_lb = new_ub = float('inf')
        else:
            # Solve model subproblem
            if config.solve_local_rnGDP:
                config.logger.debug(
                    "Screening node %s with LB %.10g and %s inactive "
                    "disjunctions."
                    % (
                        node_data.node_count,
                        node_data.obj_lb,
                        node_data.num_unbranched_disjunctions,
                    )
                )
                new_lb, new_ub = self._solve_local_rnGDP_subproblem(node_model, config)
            else:
                new_lb, new_ub = float('-inf'), float('inf')
            new_lb = max(node_data.obj_lb, new_lb)

        new_node_data = node_data._replace(
            obj_lb=new_lb, obj_ub=new_ub, is_screened=True
        )
        return new_node_data

    def _evaluate_node(self, node_data, node_model, config):
        # Solve model subproblem
        config.logger.info(
            "Exploring node %s with LB %.10g UB %.10g and %s inactive "
            "disjunctions."
            % (
                node_data.node_count,
                node_data.obj_lb,
                node_data.obj_ub,
                node_data.num_unbranched_disjunctions,
            )
        )
        new_lb, new_ub = self._solve_rnGDP_subproblem(node_model, config)
        new_node_data = node_data._replace(
            obj_lb=new_lb, obj_ub=new_ub, is_evaluated=True
        )
        return new_node_data

    def _solve_rnGDP_subproblem(self, model, config):
        subproblem = TransformationFactory('gdp.bigm').create_using(model)
        obj_sense_correction = self.objective_sense != minimize
        model_utils = model.component(self.original_util_block.name)
        subprob_utils = subproblem.component(self.original_util_block.name)

        try:
            with SuppressInfeasibleWarning():
                try:
                    fbbt(subproblem, integer_tol=config.integer_tolerance)
                except InfeasibleConstraintException:
                    # copy variable values, even if errored
                    copy_var_list_values(
                        from_list=subprob_utils.algebraic_variable_list,
                        to_list=model_utils.algebraic_variable_list,
                        config=config,
                        ignore_integrality=True,
                    )
                    return float('inf'), float('inf')
                minlp_args = dict(config.minlp_solver_args)
                if config.time_limit is not None and config.minlp_solver == 'gams':
                    elapsed = get_main_elapsed_time(self.timing)
                    remaining = max(config.time_limit - elapsed, 1)
                    minlp_args['add_options'] = minlp_args.get('add_options', [])
                    minlp_args['add_options'].append('option reslim=%s;' % remaining)
                result = SolverFactory(config.minlp_solver).solve(
                    subproblem, **minlp_args
                )
        except RuntimeError as e:
            config.logger.warning(
                "Solver encountered RuntimeError. Treating as infeasible. "
                "Msg: %s\n%s" % (str(e), traceback.format_exc())
            )
            copy_var_list_values(  # copy variable values, even if errored
                from_list=subprob_utils.algebraic_variable_list,
                to_list=model_utils.algebraic_variable_list,
                config=config,
                ignore_integrality=True,
            )
            return float('inf'), float('inf')

        term_cond = result.solver.termination_condition
        if term_cond == tc.optimal:
            assert result.solver.status is SolverStatus.ok
            lb = (
                result.problem.lower_bound
                if not obj_sense_correction
                else -result.problem.upper_bound
            )
            ub = (
                result.problem.upper_bound
                if not obj_sense_correction
                else -result.problem.lower_bound
            )
            copy_var_list_values(
                from_list=subprob_utils.algebraic_variable_list,
                to_list=model_utils.algebraic_variable_list,
                config=config,
            )
            return lb, ub
        elif term_cond == tc.locallyOptimal or term_cond == tc.feasible:
            assert result.solver.status is SolverStatus.ok
            lb = (
                result.problem.lower_bound
                if not obj_sense_correction
                else -result.problem.upper_bound
            )
            ub = (
                result.problem.upper_bound
                if not obj_sense_correction
                else -result.problem.lower_bound
            )
            # TODO handle LB absent
            copy_var_list_values(
                from_list=subprob_utils.algebraic_variable_list,
                to_list=model_utils.algebraic_variable_list,
                config=config,
            )
            return lb, ub
        elif term_cond == tc.unbounded:
            copy_var_list_values(
                from_list=subprob_utils.algebraic_variable_list,
                to_list=model_utils.algebraic_variable_list,
                config=config,
                ignore_integrality=True,
            )
            return float('-inf'), float('-inf')
        elif term_cond == tc.infeasible:
            copy_var_list_values(
                from_list=subprob_utils.algebraic_variable_list,
                to_list=model_utils.algebraic_variable_list,
                config=config,
                ignore_integrality=True,
            )
            return float('inf'), float('inf')
        else:
            config.logger.warning(
                "Unknown termination condition of %s. "
                "Treating as infeasible." % term_cond
            )
            copy_var_list_values(
                from_list=subprob_utils.algebraic_variable_list,
                to_list=model_utils.algebraic_variable_list,
                config=config,
                ignore_integrality=True,
            )
            return float('inf'), float('inf')

    def _solve_local_rnGDP_subproblem(self, model, config):
        # TODO: The returns of this method should be improved. Currently, it
        # returns trivial bounds (LB, UB) = (-inf, inf) if there is an error in
        # the solve, if the problem is infeasible, or if the problem is
        # unbounded.
        subproblem = TransformationFactory('gdp.bigm').create_using(model)
        obj_sense_correction = self.objective_sense != minimize
        subprob_utils = subproblem.component(self.original_util_block.name)
        model_utils = model.component(self.original_util_block.name)

        try:
            with SuppressInfeasibleWarning():
                result = SolverFactory(config.local_minlp_solver).solve(
                    subproblem, **config.local_minlp_solver_args
                )
        except RuntimeError as e:
            config.logger.warning(
                "Solver encountered RuntimeError. Treating as infeasible. "
                "Msg: %s\n%s" % (str(e), traceback.format_exc())
            )
            copy_var_list_values(  # copy variable values, even if errored
                from_list=subprob_utils.algebraic_variable_list,
                to_list=model_utils.algebraic_variable_list,
                config=config,
                ignore_integrality=True,
            )
            return float('-inf'), float('inf')

        term_cond = result.solver.termination_condition
        if term_cond == tc.optimal:
            assert result.solver.status is SolverStatus.ok
            ub = (
                result.problem.upper_bound
                if not obj_sense_correction
                else -result.problem.lower_bound
            )
            copy_var_list_values(
                from_list=subprob_utils.algebraic_variable_list,
                to_list=model_utils.algebraic_variable_list,
                config=config,
            )
            return float('-inf'), ub
        elif term_cond == tc.locallyOptimal or term_cond == tc.feasible:
            assert result.solver.status is SolverStatus.ok
            ub = (
                result.problem.upper_bound
                if not obj_sense_correction
                else -result.problem.lower_bound
            )
            copy_var_list_values(
                from_list=subprob_utils.algebraic_variable_list,
                to_list=model_utils.algebraic_variable_list,
                config=config,
            )
            return float('-inf'), ub
        elif term_cond == tc.unbounded:
            copy_var_list_values(
                from_list=subprob_utils.algebraic_variable_list,
                to_list=model_utils.algebraic_variable_list,
                config=config,
                ignore_integrality=True,
            )
            return float('-inf'), float('-inf')
        elif term_cond == tc.infeasible:
            copy_var_list_values(
                from_list=subprob_utils.algebraic_variable_list,
                to_list=model_utils.algebraic_variable_list,
                config=config,
                ignore_integrality=True,
            )
            # [ESJ 7/11/22]: Making this float('inf'), float('inf') doesn't
            # break tests, and makes more sense. But I'm not sure why it's not
            # that way already?
            return float('-inf'), float('inf')
        else:
            config.logger.warning(
                "Unknown termination condition of %s. "
                "Treating as infeasible." % term_cond
            )
            copy_var_list_values(
                from_list=subprob_utils.algebraic_variable_list,
                to_list=model_utils.algebraic_variable_list,
                config=config,
                ignore_integrality=True,
            )
            return float('-inf'), float('inf')
